# -*- coding: utf-8 -*-

import psycopg2

psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)

NEED_DATABASE_NAME = True
NEED_LOGIN = True
NEED_PASSWORD = True
NEED_ENCODING = True
NEED_HOST = True
NEED_PORT = True
CAN_CHANGE_TYPE = False
CAN_CHANGE_SIZE = False
UPPER_CASE = False
DDL_ROLLBACK = True
FROM = '"%s" as %s'
LEFT_OUTER_JOIN = 'left outer join "%s" as %s'
FIELD_AS = 'as'
LIKE = 'ilike'

JAM_TYPES = TEXT, INTEGER, FLOAT, CURRENCY, DATE, DATETIME, BOOLEAN, BLOB = range(1, 9)
FIELD_TYPES = {
    INTEGER: 'INTEGER',
    TEXT: 'VARCHAR',
    FLOAT: 'NUMERIC',
    CURRENCY: 'NUMERIC',
    DATE: 'DATE',
    DATETIME: 'TIMESTAMP',
    BOOLEAN: 'INTEGER',
    BLOB: 'BYTEA'
}

def connect(database, user, password, host, port, encoding):
    return psycopg2.connect(database=database, user=user, password=password, host=host, port=port)

def get_lastrowid(cursor):
    return None

def get_select(query, start, end, fields):
    offset = query['__offset']
    limit = query['__limit']
    result = 'select %s from %s' % (start, end)
    if limit:
        result += ' limit %d offset %d' % (limit, offset)
    return result

def process_sql_params(params, cursor):
    result = []
    for p in params:
        if type(p) == tuple:
            value, data_type = p
            if data_type == BLOB:
                if type(value) == unicode:
                    value = value.encode('utf-8')
                value = psycopg2.Binary(value)
        else:
            value = p
        result.append(value)
    return result

def process_sql_result(rows):
    result = []
    for row in rows:
        fields = []
        for field in row:
            if type(field) == buffer:
                field = str(field)
            fields.append(field)
        result.append(fields)
    return result

def cast_date(date_str):
    return "cast('" + date_str + "' as date)"

def cast_datetime(datetime_str):
    return "cast('" + datetime_str + "' as timestamp)"

def value_literal(index):
    return '%s'

def upper_function():
    pass

def create_table_sql(table_name, fields, foreign_fields=None):
    result = []
    primary_key = ''
    seq_name = '%s_id_seq' % table_name
    result.append(set_case('create sequence "%s"' % seq_name))
    sql = 'create table "%s"\n(\n' % set_case(table_name)
    for field in fields:
        if field['primary_key']:
            primary_key = set_case(field['field_name'])
            sql += set_case('"%s" %s primary key default nextval(\'"%s"\')' % \
                (field['field_name'], FIELD_TYPES[field['data_type']], seq_name))
        else:
            sql += set_case('"%s" %s' % (field['field_name'], FIELD_TYPES[field['data_type']]))
        if field['size'] != 0 and field['data_type'] == TEXT:
            sql += '(%d)' % field['size']
        if field['default_value'] and not field['primary_key']:
            if field['data_type'] == TEXT:
                sql += set_case(" default '%s'" % field['default_value'])
            else:
                sql += set_case(' default %s' % field['default_value'])
        sql +=  ',\n'
    sql = sql[:-2]
    sql += ')\n'
    result.append(sql)
    result.append(set_case('alter sequence "%s" owned by "%s"."%s"' % (seq_name, table_name, primary_key)))
    return result

def delete_table_sql(table_name):
    result = []
    result.append(set_case('drop table "%s"' % table_name))
    return result

def create_index_sql(index_name, table_name, unique, fields, desc):
    return set_case('create %s index "%s" on "%s" (%s)' % \
        (unique, index_name, table_name, fields))

def create_foreign_index_sql(table_name, index_name, key, ref, primary_key):
    return set_case('alter table %s add constraint %s foreign key (%s) references %s(%s) match simple' % \
        (table_name, index_name, key, ref, primary_key))

def delete_index(table_name, index_name):
    return set_case('drop index "%s"' % index_name)

def delete_foreign_index(table_name, index_name):
    return set_case('alter table %s drop constraint %s' % (table_name, index_name))

def add_field_sql(table_name, field):
    result = set_case('alter table "%s" add column "%s" %s' %
        (table_name, field['field_name'], FIELD_TYPES[field['data_type']]))
    if field['size']:
        result += '(%d)' % field['size']
    if field['default_value']:
        if field['data_type'] == TEXT:
            result += " default '%s'" % set_case(field['default_value'])
        else:
            result += ' default %s' % field['default_value']
    return result

def del_field_sql(table_name, field):
    return set_case('alter table "%s" drop column "%s"' % (table_name, field['field_name']))

def change_field_sql(table_name, old_field, new_field):
    result = []
    if FIELD_TYPES[old_field['data_type']] != FIELD_TYPES[new_field['data_type']] \
        or old_field['size'] != new_field['size']:
        raise Exception, u"Don't know how to change field's size or type of %s" % old_field['field_name']
    if old_field['field_name'] != new_field['field_name']:
        result.append(set_case('alter table "%s" rename column  "%s" TO "%s"' % \
            (table_name, old_field['field_name'], new_field['field_name'])))
    if old_field['default_value'] != new_field['default_value']:
        if new_field['default_value']:
            if new_field['data_type'] == TEXT:
                sql = set_case('alter table "%s" alter "%s" set default' % \
                    (table_name, new_field['field_name']))
                sql +=  " '%s'" % new_field['default_value']
            else:
                sql = set_case('alter table "%s" alter "%s" set default %s' % \
                    (table_name, new_field['field_name'], new_field['default_value']))
        else:
            sql = set_case('alter table "%s" alter "%s" drop default' % \
                (table_name, new_field['field_name']))
        result.append(sql)
    return result


def set_case(string):
    return string.lower()

def get_sequence_name(table_name):
    return set_case('%s_id_seq' % table_name)

def next_sequence_value_sql(table_name):
    return set_case("select nextval('%s')" % get_sequence_name(table_name))

def restart_sequence_sql(table_name, value):
    return set_case('alter sequence %s restart with %d' % (get_sequence_name(table_name), value))




lang_dict = {
    'af': 'Afrikaans',
    'sq': 'Albanian',
    'ar-sa': 'Arabic (Saudi Arabia)',
    'ar-iq': 'Arabic (Iraq)',
    'ar-eg': 'Arabic (Egypt)',
    'ar-ly': 'Arabic (Libya)',
    'ar-dz': 'Arabic (Algeria)',
    'ar-ma': 'Arabic (Morocco)',
    'ar-tn': 'Arabic (Tunisia)',
    'ar-om': 'Arabic (Oman)',
    'ar-ye': 'Arabic (Yemen)',
    'ar-sy': 'Arabic (Syria)',
    'ar-jo': 'Arabic (Jordan)',
    'ar-lb': 'Arabic (Lebanon)',
    'ar-kw': 'Arabic (Kuwait)',
    'ar-ae': 'Arabic (U.A.E.)',
    'ar-bh': 'Arabic (Bahrain)',
    'ar-qa': 'Arabic (Qatar)',
    'eu': 'Basque',
    'bg': 'Bulgarian',
    'be': 'Belarusian',
    'ca': 'Catalan',
    'zh-tw': 'Chinese (Taiwan)',
    'zh-cn': 'Chinese (PRC)',
    'zh-hk': 'Chinese (Hong Kong SAR)',
    'zh-sg': 'Chinese (Singapore)',
    'hr': 'Croatian',
    'cs': 'Czech',
    'da': 'Danish',
    'nl': 'Dutch (Standard)',
    'nl-be': 'Dutch (Belgium)',
    'en': 'English',
    'en-us': 'English (United States)',
    'en-gb': 'English (United Kingdom)',
    'en-au': 'English (Australia)',
    'en-ca': 'English (Canada)',
    'en-nz': 'English (New Zealand)',
    'en-ie': 'English (Ireland)',
    'en-za': 'English (South Africa)',
    'en-jm': 'English (Jamaica)',
    'en-029': 'English (Caribbean)',
    'en-bz': 'English (Belize)',
    'en-tt': 'English (Trinidad)',
    'et': 'Estonian',
    'fo': 'Faeroese',
    'fa': 'Farsi',
    'fi': 'Finnish',
    'fr': 'French (Standard)',
    'fr-be': 'French (Belgium)',
    'fr-ca': 'French (Canada)',
    'fr-ch': 'French (Switzerland)',
    'fr-lu': 'French (Luxembourg)',
    'gd': 'Gaelic (Scotland)',
    'gd-ie': 'Gaelic (Ireland)',
    'de': 'German (Standard)',
    'de-ch': 'German (Switzerland)',
    'de-at': 'German (Austria)',
    'de-lu': 'German (Luxembourg)',
    'de-li': 'German (Liechtenstein)',
    'el': 'Greek',
    'he': 'Hebrew',
    'hi': 'Hindi',
    'hu': 'Hungarian',
    'is': 'Icelandic',
    'id': 'Indonesian',
    'it': 'Italian (Standard)',
    'it-ch': 'Italian (Switzerland)',
    'ja': 'Japanese',
    'ko': 'Korean',
    'ko': 'Korean (Johab)',
    'lv': 'Latvian',
    'lt': 'Lithuanian',
    'mk': 'FYRO Macedonian ms Malaysian',
    'mt': 'Maltese',
    'no': 'Norwegian (Bokmal)',
    'no': 'Norwegian (Nynorsk)',
    'pl': 'Polish',
    'pt-br': 'Portuguese (Brazil)',
    'pt': 'Portuguese (Portugal)',
    'rm': 'Rhaeto-Romanic',
    'ro': 'Romanian',
    'ro-mo': 'Romanian (Moldavia)',
    'ru': 'Russian',
    'ru-mo': 'Russian (Moldavia)',
    'sz': 'Sami (Lappish)',
    'sr': 'Serbian (Cyrillic)',
    'sr': 'Serbian (Latin)',
    'sk': 'Slovak',
    'sl': 'Slovenian',
    'sb': 'Sorbian',
    'es': 'Spanish (Spain, traditional)',
    'es': 'Spanish (Spain, modern)',
    'es-ar': 'Spanish (Argentina)',
    'es-gt': 'Spanish (Guatemala)',
    'es-cr': 'Spanish (Costa Rica)',
    'es-pa': 'Spanish (Panama)',
    'es-do': 'Spanish (Dominican Republic)',
    'es-mx': 'Spanish (Mexico)',
    'es-ve': 'Spanish (Venezuela)',
    'es-co': 'Spanish (Colombia)',
    'es-pe': 'Spanish (Peru)',
    'es-ec': 'Spanish (Ecuador)',
    'es-cl': 'Spanish (Chile)',
    'es-uy': 'Spanish (Uruguay)',
    'es-py': 'Spanish (Paraguay)',
    'es-bo': 'Spanish (Bolivia)',
    'es-sv': 'Spanish (El Salvador)',
    'es-hn': 'Spanish (Honduras)',
    'es-ni': 'Spanish (Nicaragua)',
    'es-pr': 'Spanish (Puerto Rico)',
    'sx': 'Sutu',
    'sv': 'Swedish',
    'sv-fi': 'Swedish (Finland)',
    'th': 'Thai',
    'ts': 'Tsonga',
    'tn': 'Tswana',
    'tr': 'Turkish',
    'uk': 'Ukrainian',
    'ur': 'Urdu',
    've': 'Venda',
    'vi': 'Vietnamese',
    'xh': 'Xhosa',
    'ji': 'Yiddish',
    'zu': 'Zulu'
}



# -*- coding: utf-8 -*-

ENG, RUS, HUN, POR = range(1, 5)
LANGUAGE = ('English', 'Russian', 'Hungarian', 'Portuguese')

def get_lang_dict(language):
    if language == ENG:
        import english as lang
    elif language == RUS:
        import russian as lang
    elif language == HUN:
        import hungarian as lang
    elif language == POR:
        import portuguese as lang
    return lang.dictionary



# -*- coding: utf-8 -*-

import sys
import os
import json
import uuid
import traceback
import datetime
from threading import Lock
from types import MethodType
import jam

sys.path.insert(1, os.path.join(os.path.dirname(jam.__file__), 'third_party'))

from werkzeug.wrappers import Request, Response
from werkzeug.routing import Map, Rule
from werkzeug.exceptions import HTTPException, NotFound
from werkzeug.wsgi import SharedDataMiddleware, peek_path_info, get_path_info

import common
from items import AbortException

def create_application(from_file):
    if from_file:
        work_dir = os.path.dirname(os.path.abspath(from_file))
    else:
        work_dir = os.getcwd()
    os.chdir(work_dir)
    static_files = {
        '/static':  os.path.join(work_dir, 'static')
    }
    application = App(work_dir)
    return SharedDataMiddleware(application, static_files)

class App():
    def __init__(self, work_dir):
        self.started = datetime.datetime.now()
        self.work_dir = work_dir
        self._loading = False
        self._load_lock = Lock()
        self.admin = None
        self.task = None
        self._busy = 0
        self.pid = os.getpid()
        self.task_server_modified = False
        self.task_client_modified = True
        self.under_maintenance = False
        self.jam_dir = os.path.dirname(jam.__file__)
        self.jam_version = jam.version()
        self.application_files = {
            '/': self.work_dir,
            '/jam/': self.jam_dir
        }
        self.fileserver = SharedDataMiddleware(None, self.application_files, cache_timeout=1)
        self.url_map = Map([
            Rule('/', endpoint='root_file'),
            Rule('/<file_name>', endpoint='root_file'),
            Rule('/js/<file_name>', endpoint='file'),
            Rule('/css/<file_name>', endpoint='file'),
            Rule('/jam/js/<file_name>', endpoint='file'),
            Rule('/jam/js/ace/<file_name>', endpoint='file'),
            Rule('/jam/css/<file_name>', endpoint='file'),
            Rule('/jam/img/<file_name>', endpoint='file'),
            Rule('/api', endpoint='api'),
            Rule('/upload', endpoint='upload')
        ])
        self.admin = self.create_admin()

    def create_admin(self):
        from adm_server import create_admin
        return create_admin(self)

    def get_task(self):
        if self.task:
            return self.task
        else:
            if not self._loading:
                self._loading = True
                try:
                    with self._load_lock:
                        self.task = self.admin.create_task()
                finally:
                    self._loading = False
            return self.task

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def wsgi_app(self, environ, start_response):
        request = Request(environ)
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            if endpoint in ['file', 'root_file']:
                return self.serve_file(environ, start_response, endpoint, **values)
            elif endpoint in ['api', 'upload']:
                response = getattr(self, 'on_' + endpoint)(request, **values)
        except HTTPException, e:
            if peek_path_info(environ) == 'ext':
                response = self.on_ext(request)
            else:
                response = e
        return response(environ, start_response)

    def set_no_cache_headers(self, response):
        response.headers['Content-Type'] = 'text/html'
        response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'
        response.headers['Pragma'] = 'no-cache'
        response.headers['Expires'] = 0

    def serve_himself(self, environ, start_response, file_name):
        response = Response()
        self.set_no_cache_headers(response)
        if file_name == 'admin.html':
            file_name = os.path.join(self.jam_dir, file_name)
        with open(file_name, 'r') as f:
            response.set_data(f.read())
        return response(environ, start_response)

    def serve_file(self, environ, start_response, endpoint, file_name=None):
        if endpoint == 'root_file':
            if not file_name:
                file_name = 'index.html'
                environ['PATH_INFO'] = environ['PATH_INFO'] + '/index.html'
            if file_name == 'index.html':
                if self.get_task():
                    self.check_task_client_modified()
                    return self.serve_himself(environ, start_response, file_name)
                else:
                    response = Response(self.admin.lang['no_task'])
                    self.set_no_cache_headers(response)
                    return response(environ, start_response)
            elif file_name == 'admin.html':
                return self.serve_himself(environ, start_response, file_name)
        try:
            if file_name:
                base, ext = os.path.splitext(file_name)
            if common.SETTINGS['COMPRESSED_JS'] and ext and ext in ['.js', '.css']:
                try:
                    cur_path_info = environ['PATH_INFO']
                    min_file_name = base + '.min' + ext
                    environ['PATH_INFO'] = environ['PATH_INFO'].replace(file_name, min_file_name)
                    return self.fileserver(environ, start_response)
                except:
                    environ['PATH_INFO'] = cur_path_info
                    return self.fileserver(environ, start_response)
            else:
                return self.fileserver(environ, start_response)
        except Exception, e:
            return Response('')(environ, start_response)


    def create_post_response(self, request, result):
        response = Response()
        accepts_gzip = 0
        try:
            if request.environ.get("HTTP_ACCEPT_ENCODING").find("gzip") != -1:
                accepts_gzip = 1
        except:
            pass
        buff = json.dumps(result, default=common.json_defaul_handler)
        response.headers['Content-Type'] = 'application/json'
        if accepts_gzip:
            buff = common.compressBuf(buff)
            response.headers['Content-encoding'] = 'gzip'
            response.headers['Content-Length'] = str(len(buff))
        response.set_data(buff)
        return response

    def on_api(self, request):
        if request.method == 'POST':
            r = {'result': None, 'error': None}
            try:
                method, user_id, task_id, item_id, params, date = json.loads(request.get_data())
                r['result'] = self.process_request(request.environ, method, user_id, task_id, item_id, params)
            except AbortException, e:
                print traceback.format_exc()
                r['result'] = {'data': [None, e.message]}
                r['error'] = e.message
            except Exception, e:
                print traceback.format_exc()
                if common.SETTINGS['DEBUGGING'] and task_id != 0:
                    raise
                r['result'] = {'data': [None, e.message]}
                r['error'] = e.message
            return self.create_post_response(request, r)

    def process_request(self, env, method, user_uuid=None, task_id=None, item_id=None, params=None, ext=None):
        user_info = {}
        is_admin = task_id == 0
        if is_admin:
            task = self.admin
        else:
            task = self.get_task()
        if not task:
            return {'status': common.NO_PROJECT, 'data': None}
        elif self.under_maintenance:
            return {'status': common.UNDER_MAINTAINANCE, 'data': None}
        elif method == 'login':
            return {'status': common.RESPONSE, 'data': self.admin.login(params[0], params[1], is_admin, env)}
        elif method == 'logout':
            self.admin.logout(params, is_admin, env)
            return {'status': common.NOT_LOGGED, 'data': common.NOT_LOGGED}
        if ext:
            obj = task
        else:
            if self.admin.safe_mode:
                user_info = self.admin.get_user_info(user_uuid, is_admin, env)
                if not user_info:
                    return {'status': common.NOT_LOGGED, 'data': common.NOT_LOGGED}
            elif not user_uuid is None:
                return {'status': common.NOT_LOGGED, 'data': common.NOT_LOGGED}
            obj = task
            if task:
                obj = task.item_by_ID(item_id)
        self._busy += 1
        try:
            data = None
            if task.on_request:
                data = task.on_request(task, user_info, env, method, obj, params, ext)
            if not data:
                data = self.get_response(is_admin, env, method, user_info, task_id, obj, params, ext)
        finally:
            self._busy -= 1
        return {'status': common.RESPONSE, 'data': data, 'version': task.version, 'server_date': self.started}

    def get_response(self, is_admin, env, method, user_info, task_id, item, params, ext):
        if ext:
            if item.on_ext_request:
                return item.on_ext_request(item, method, params, env)
        elif method == 'server_function':
            return self.server_func(item, params[0], params[1], env)
        elif method == 'open':
            if self.admin.has_privilege(user_info, item, 'can_view'):
                return item.select_records(params, user_info, env)
            else:
                return [], item.task.lang['cant_view'] % item.item_caption
        elif method == 'get_record_count':
            return item.get_record_count(params, env)
        elif method == 'apply_changes':
            return item.apply_changes(params, self.admin.find_privileges(user_info, item), user_info, env)
        elif method == 'print_report':
            url = None
            error = None
            if self.admin.has_privilege(user_info, item, 'can_view'):
                url = item.print_report(*params)
            else:
                error = item.task.lang['cant_view'] % item.item_caption
            return url, error
        elif method == 'init_client':
            return self.init_client(user_info, is_admin)

    def server_func(self, obj, func_name, params, env):
        result = None
        error = ''
        func = getattr(obj, func_name)
        if func:
            if func_name[-4:] == '_env':
                params = list(params)
                params.append(env)
            result = func(obj, *params)
            #~ if isinstance(func, MethodType):
                #~ result = func(*params)
            #~ else:
                #~ result = func(obj, *params)
        else:
            raise Exception, 'item: %s no server function with name %s' % (obj.item_name, func_name)
        return result, error

    def check_task_server_modified(self):
        if self.task_server_modified:
            self.admin.reload_task()
            self.task_server_modified = False

    def check_task_client_modified(self):
        if self.task_client_modified:
            self.admin.update_events_code()
            self.task_client_modified = False

    def init_client(self, user_info, is_admin):
        if is_admin:
            task = self.admin
        else:
            task = self.task
            self.check_task_server_modified()
        if user_info:
            priv = self.admin.get_privileges(user_info['role_id'])
        else:
            priv = None
        result = {
            'task': task.get_info(),
            'settings': self.admin.get_settings(),
            'language': self.admin.lang,
            'user_info': user_info,
            'privileges': priv,
            'demo': jam.common.DEMO
        }
        return result, ''

    def on_ext(self, request):
        if request.method == 'POST':
            r = {'result': None, 'error': None}
            method = get_path_info(request.environ)
            params = json.loads(request.get_data())
            user_id = None
            task_id = None
            item_id = None
            ext = True
            try:
                r['result'] = self.process_request(request.environ,
                    method, user_id, task_id, item_id, params, ext)
            except AbortException, e:
                print traceback.format_exc()
                r['result'] = {'data': [None, e.message]}
                r['error'] = e.message
            except Exception, e:
                print traceback.format_exc()
                if common.SETTINGS['DEBUGGING'] and task_id != 0:
                    raise
                r['result'] = {'data': [None, e.message]}
                r['error'] = e.message
            return self.create_post_response(request, r)

    def on_upload(self, request):

        def find_param(data):
            pos = data.find(';')
            return data[:pos], pos + 1

        def read_user_info(data):
            info_len, pos = find_param(data)
            info_len = int(info_len)
            user_info = data[pos:pos+info_len]
            task_ID, p = find_param(user_info)
            user_id = user_info[p:]
            pos = pos + info_len + 1
            return user_id, int(task_ID), pos

        if request.method == 'POST':
            try:
                data = request.get_data()
                header = []
                header_str = ''
                length = 0
                string = ''
                user_id, task_ID, pos = read_user_info(data)
                if self.admin.safe_mode:
                    user_info = self.admin.get_user_info(user_id, task_ID == 0, request.environ)
                    if not user_info:
                        return Response()
                for s in data[pos:]:
                    header_str += s
                    if s == ';':
                        if len(header) == 0:
                            length = int(string)
                        header.append(int(string))
                        if len(header) == 2 * (length + 1):
                            break;
                        string = ''
                    else:
                        string += s
                start = len(header_str) + pos
                path = os.path.join(os.getcwd(), os.path.normpath(data[start: start + header[1]]))
                if not os.path.exists(path):
                    os.makedirs(path)
                start = start + header[1]
                for i in range(length):
                    index = 2 * i + 2
                    file_name = data[start: start + header[index]]
                    start = start + header[index]
                    index += 1
                    content = data[start: start + header[index]]
                    file_name = os.path.join(path, file_name)
                    with open(file_name, 'wb') as f:
                        f.write(content)
                    os.chmod(file_name, 0o666)
                    start = start + header[index]
            except:
                print traceback.format_exc()
            return Response()

    def stop(self, sigvalue):
        self.kill()

    def kill(self):
        import signal, subprocess
        if os.name == "nt":
            subprocess.Popen("taskkill /F /T /pid %i" % self.pid, shell=True)
        else :
            os.killpg(self.pid, signal.SIGKILL)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

def process_html(file_name):
    result = []
    with open(file_name) as f:
        content = f.readlines()
    for i, line in enumerate(content):
        cur_line = line.strip()
        if cur_line == '<div class="sphinxsidebarwrapper">':
            if content[i + 1].find('Table Of Contents') != -1:
                content[i + 1] = '<h3>Contents</h3>\n'
        if cur_line == '<h3>This Page</h3>':
            for j in range(6):
                content[i + j - 1] = ''
        if cur_line.find('<table') != -1 and cur_line.find('class="docutils"') != -1:
            content[i] = line.replace('class="docutils"', 'class="table-condensed table-bordered table-striped"')
    for line in content:
        if line:
            result.append(line)
    with open(file_name, 'w') as f:
        f.write(''.join(result))

def update_html(folder):
    for root, dirs, files in os.walk(folder):
        for name in files:
            n, ext = os.path.splitext(name)
            if ext == '.html':
                process_html(os.path.join(root, name))

if __name__ == '__main__':
    folder = os.path.join(os.getcwd(), '_build', 'html')
    update_html(folder)



# -*- coding: utf-8 -*-

from __future__ import with_statement
import sys, os

import Queue
import multiprocessing
import threading
import zipfile
from xml.dom.minidom import parseString
from xml.sax.saxutils import escape
import uuid
import datetime, time
import traceback
import inspect
import json

import common, db.db_modules as db_modules
from items import *
from dataset import *
from sql import *
import adm_server

class ServerDataset(Dataset, SQL):
    def __init__(self, table_name='', soft_delete=True):
        Dataset.__init__(self)
        self.ID = None
        self.table_name = table_name
        self._order_by = []
        self.values = None
        self.on_open = None
        self.on_apply = None
        self.on_count = None
        self.on_get_field_text = None
        self.deleted_field_name = None
        self.soft_delete = soft_delete
        self.virtual_table = False

    def copy(self, filters=True, details=True, handlers=True):
        if self.master:
            raise DatasetException(u'A detail item can not be copied: %s' % self.item_name)
        return self._copy(filters, details, handlers)

    def _copy(self, filters=True, details=True, handlers=True):
        result = super(ServerDataset, self)._copy(filters, details, handlers)
        result.table_name = self.table_name
        result._order_by = self._order_by
        result.soft_delete = self.soft_delete
        result._primary_key = self._primary_key
        result._deleted_flag = self._deleted_flag
        result._master_id = self._master_id
        result._master_rec_id = self._master_rec_id
        return result

    def get_event(self, caption):
        return getattr(caption)

    def add_field(self, field_id, field_name, field_caption, data_type, required = False,
        item = None, object_field = None, visible = True, index=0, edit_visible = True, edit_index = 0, read_only = False,
        expand = False, word_wrap = False, size = 0, default_value=None, default = False, calculated = False, editable = False,
        master_field = None, alignment=None, lookup_values=None, enable_typeahead=False, field_help=None, field_placeholder=None):

        field_def = self.add_field_def(field_id, field_name, field_caption, data_type, required, item, object_field, visible,
            index, edit_visible, edit_index, read_only, expand, word_wrap, size, default_value, default, calculated, editable,
            master_field, alignment, lookup_values, enable_typeahead, field_help, field_placeholder)
        field = DBField(self, field_def)
        self._fields.append(field)
        return field

    def add_filter(self, name, caption, field_name, filter_type = common.FILTER_EQ,
        data_type = None, visible = True, filter_help=None, filter_placeholder=None):

        filter_def = self.add_filter_def(name, caption, field_name, filter_type,
            data_type, visible, filter_help, filter_placeholder)
        fltr = DBFilter(self, filter_def)
        self.filters.append(fltr)
        return fltr

    def do_internal_open(self, params):
        return self.select_records(params)

    def do_apply(self, params=None):
        if not self.master and self.log_changes:
            if self.item_state != common.STATE_BROWSE:
                raise Exception, u'Item: %s is not in browse state. Apply requires browse state.'
            changes = {}
            self.change_log.get_changes(changes)
            if changes['data']:
                data, error = self.apply_changes((changes, params),
                    {'can_view': True, 'can_create': True, 'can_edit': True, 'can_delete': True})
                if error:
                    raise Exception, error
                else:
                    self.change_log.update(data)

    def add_detail(self, table):
        detail = Detail(self, table.item_name, table.item_caption, table.table_name)
        self.details.append(detail)
        detail.owner = self
        detail.init_fields()
        return detail

    def detail_by_name(self, caption):
        for table in self.details:
            if table.item_name == caption:
                return table

    def change_order(self, *fields):
        self._order_by = []
        for field in fields:
            field_name = field
            desc = False
            if field[0] == '-':
                desc = True
                field_name = field[1:]
            try:
                fld = self._field_by_name(field_name)
            except:
                raise RuntimeError('%s: change_order method arument error - %s' % (self.item_name, field))
            self._order_by.append([fld.ID, desc])
        return self

    def get_record_count(self, params, user_info=None, enviroment=None):
        result = 0;
        if self.on_count:
            result, error_mes = self.on_count(self, params, user_info, enviroment)
        else:
            error_mes = ''
            result = 0
            sql = self.get_record_count_query(params)
            rows = self.task.execute_select(sql)
            result = rows[0][0]
        return result, error_mes

    def select_records(self, params, user_info=None, enviroment=None):
        if self.on_open:
            rows, error_mes = self.on_open(self, params, user_info, enviroment)
        else:
            sql = self.get_select_statement(params)
            error_mes = ''
            rows = []
            rows = self.task.execute_select(sql)
        return rows, error_mes

    def apply_delta(self, delta, privileges=None, user_info=None):
        sql = delta.apply_sql(user_info, privileges)
        return self.task.execute(sql)

    def apply_changes(self, data, privileges, user_info=None, enviroment=None):
        if common.DEMO:
            raise Exception, u'Demo version - changes are not allowed.'
        error = None
        result = None
        changes, params = data
        if not params:
            params = {}
        delta = self.delta(changes)
        if self.on_apply:
            result, error = self.on_apply(self, delta, params, privileges, user_info, enviroment)
        else:
            result, error = self.apply_delta(delta, privileges, user_info)
        return result, error

    def update_deleted(self):
        if self._is_delta and len(self.details):
            rec_no = self.rec_no
            try:
                for it in self:
                    if it.rec_deleted():
                        for detail in self.details:
                            fields = []
                            for field in detail.fields:
                                fields.append(field.field_name)
                            det = self.task.item_by_name(detail.item_name).copy()
                            where = {
                                det._master_id: self.ID,
                                det._master_rec_id: self._primary_key_field.value
                            }
                            det.open(fields=fields, expanded=detail.expanded, where=where)
                            it.edit()
                            for d in det:
                                detail.append()
                                for field in detail.fields:
                                    f = det.field_by_name(field.field_name)
                                    field.set_value(f.value, f.lookup_value)
                                detail.post()
                            it.post()
                            for d in detail:
                                d.record_status = common.RECORD_DELETED
            finally:
                self.rec_no = rec_no

    def field_by_id(self, id_value, field_name):
        return self.get_field_by_id((id_value, field_name))

    def get_field_by_id(self, params):
        id_value, fields = params
        if not (isinstance(fields, tuple) or isinstance(fields, list)):
            fields = [fields]
        copy = self.copy()
        copy.set_where(id=id_value)
        copy.open(fields=fields)
        if copy.record_count() == 1:
            result = []
            for field_name in fields:
                result.append(copy.field_by_name(field_name).value)
            if len(fields) == 1:
                return result[0]
            else:
                return result
        return

class Item(AbstrItem, ServerDataset):
    def __init__(self, owner, name, caption, visible = True,
            table_name='', view_template='', js_filename='', soft_delete=True):
        AbstrItem.__init__(self, owner, name, caption, visible, js_filename=js_filename)
        ServerDataset.__init__(self, table_name, soft_delete)
        self.item_type_id = None
        self.reports = []

    def get_reports_info(self):
        result = []
        for report in self.reports:
            result.append(report.ID)
        return result


class Param(DBField):
    def __init__(self, owner, param_def):
        DBField.__init__(self, owner, param_def)
        self.field_kind = common.PARAM_FIELD
        if self.data_type == common.TEXT:
            self.field_size = 1000
        else:
            self.field_size = 0
        self.param_name = self.field_name
        self.param_caption = self.field_caption
        self._value = None
        self._lookup_value = None
        setattr(owner, self.param_name, self)


    def system_field(self):
        return False

    def get_data(self):
        return self._value

    def set_data(self, value):
        self._value = value

    def get_lookup_data(self):
        return self._lookup_value

    def set_lookup_data(self, value):
        self._lookup_value = value

    def _do_before_changed(self):
        pass

    def _change_lookup_field(self, lookup_value=None, slave_field_values=None):
        pass

    def raw_display_text(self):
        result = ''
        if self.lookup_item:
            result = self.lookup_text
        else:
            result = self.text
        return result

    def copy(self, owner):
        result = Param(owner, self.param_caption, self.field_name, self.data_type,
            self.lookup_item, self.lookup_field, self.required,
            self.edit_visible, self.alignment)
        return result


class Report(AbstrReport):
    def __init__(self, owner, name='', caption='', visible = True,
            table_name='', view_template='', js_filename=''):
        AbstrReport.__init__(self, owner, name, caption, visible, js_filename=js_filename)
        self.param_defs = []
        self.params = []
        self.template = view_template
        self.template_name = None
        self.template_content = {}

        self.on_before_generate = None
        self.on_generate = None
        self.on_after_generate = None
        self.on_parsed = None
        self.on_before_save_report = None

        self.on_before_append = None
        self.on_after_append = None
        self.on_before_edit = None
        self.on_after_edit = None
        self.on_before_open = None
        self.on_after_open = None
        self.on_before_post = None
        self.on_after_post = None
        self.on_before_delete = None
        self.on_after_delete = None
        self.on_before_cancel = None
        self.on_after_cancel = None
        self.on_before_apply = None
        self.on_after_apply = None
        self.on_before_scroll = None
        self.on_after_scroll = None
        self.on_filter_record = None
        self.on_field_changed = None
        self.on_filters_applied = None
        self.on_before_field_changed = None
        self.on_filter_value_changed = None
        self.on_field_validate = None
        self.on_get_field_text = None


    def add_param(self, caption='', name='', data_type=common.INTEGER,
        obj=None, obj_field=None, required=True, visible=True, alignment=None,
        enable_typeahead=None, lookup_values=None, param_help=None, param_placeholder=None):
        param_def = self.add_param_def(caption, name, data_type, obj,
            obj_field, required, visible, alignment, enable_typeahead,
            lookup_values, param_help, param_placeholder)
        param = Param(self, param_def)
        self.params.append(param)

    def add_param_def(self, param_caption='', param_name='', data_type=common.INTEGER,
        lookup_item=None, lookup_field=None, required=True, visible=True,
        alignment=0, enable_typeahead=False, lookup_values=None, param_help=None,
        param_placeholder=None):
        param_def = [None for i in range(len(FIELD_DEF))]
        param_def[FIELD_NAME] = param_name
        param_def[NAME] = param_caption
        param_def[FIELD_DATA_TYPE] = data_type
        param_def[REQUIRED] = required
        param_def[LOOKUP_ITEM] = lookup_item
        param_def[LOOKUP_FIELD] = lookup_field
        param_def[FIELD_EDIT_VISIBLE] = visible
        param_def[FIELD_ALIGNMENT] = alignment
        param_def[FIELD_ENABLE_TYPEAHEAD] = enable_typeahead
        param_def[FIELD_LOOKUP_VALUES] = lookup_values
        param_def[FIELD_HELP] = param_help
        param_def[FIELD_PLACEHOLDER] = param_placeholder
        self.param_defs.append(param_def)
        return param_def

    def prepare_params(self):
        for param in self.params:
            if param.lookup_item and type(param.lookup_item) == int:
                param.lookup_item = self.task.item_by_ID(param.lookup_item)
            if param.lookup_field and type(param.lookup_field) == int:
                param.lookup_field = param.lookup_item._field_by_ID(param.lookup_field).field_name

    def copy(self):
        result = self.__class__(self.owner, self.item_name, self.item_caption, self.visible,
            '', self.template, '');
        result.on_before_generate = self.on_before_generate
        result.on_generate = self.on_generate
        result.on_after_generate = self.on_after_generate
        result.on_before_save_report = self.on_before_save_report
        result.on_parsed = self.on_parsed
        result.param_defs = self.param_defs
        result.template_content = self.template_content.copy()
        result.template_name = self.template_name
        for param_def in result.param_defs:
            param = Param(result, param_def)
            result.params.append(param)
        result.prepare_params()
        return  result

    def print_report(self, param_values, url, ext=None):
        if not self.template_content:
            self.parse_template()
        copy_report = self.copy()
        return copy_report.generate(param_values, url, ext)

    def generate_file_name(self, ext=None):
        if not ext:
            ext = 'ods'
        file_name = self.item_name + '_' + datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S.%f') + '.' + ext
        #~ if os.name == "nt":
            #~ file_name = self.item_name + '_' + datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S.%f') + '.' + ext
        #~ else:
            #~ file_name = self.item_caption + '_' + datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S.%f') + '.' + ext
        file_name = escape(file_name, {':': '-', '/': '_', '\\': '_'})
        return os.path.abspath(os.path.join(self.task.work_dir, 'static', 'reports', file_name))

    def generate(self, param_values, url, ext):
        self.extension = ext
        self.url = url
        template = self.template
        for i, param in enumerate(self.params):
            param.value = param_values[i]
        if self.on_before_generate:
            self.on_before_generate(self)
        if template != self.template:
            self.template_content = None
        if self.template:
            if not self.template_content:
                self.parse_template()
            if self.on_parsed:
                self.on_parsed(self)
            self.content_name = os.path.join(self.task.work_dir, 'reports', 'content%s.xml' % time.time())
            self.content = open(self.content_name, 'wb')
            try:
                self.report_filename = self.generate_file_name()
                file_name = os.path.basename(self.report_filename)
                static_dir = os.path.dirname(self.report_filename)
                if not os.path.exists(static_dir):
                    os.makedirs(static_dir)
                self.content.write(self.template_content['header'])
                self.content.write(self.template_content['columns'])
                self.content.write(self.template_content['rows'])
                if self.on_generate:
                    self.on_generate(self)
                self.content.write(self.template_content['footer'])
                self.save()
            finally:
                if not self.content.closed:
                    self.content.close()
                if os.path.exists(self.content_name):
                    os.remove(self.content_name)
            if ext and (ext != 'ods'):
                converted = False
                if self.owner.on_convert_report:
                    try:
                        self.owner.on_convert_report(self)
                        converted = True
                    except:
                        pass
                #~ if not converted:
                    #~ # OpenOffice must be running in server mode
                    #~ # soffice --headless --accept="socket,host=127.0.0.1,port=2002;urp;"
                    #~ ext_file_name = self.report_filename.replace('.ods', '.' + ext)
                    #~ try:
                        #~ from DocumentConverter import DocumentConverter
                        #~ converter = DocumentConverter()
                        #~ converter.convert(self.report_filename, ext_file_name)
                        #~ converted = True
                    #~ except:
                        #~ pass
                if not converted:
                    converted = self.task.convert_report(self, ext)
                converted_file = self.report_filename.replace('.ods', '.' + ext)
                if converted and os.path.exists(converted_file):
                    self.delete_report(self.report_filename)
                    file_name = file_name.replace('.ods', '.' + ext)
            self.report_filename = os.path.join(self.task.work_dir, 'static', 'reports', file_name)
            self.report_url = self.report_filename
            if self.url:
                self.report_url = os.path.join(self.url, 'static', 'reports', file_name)
        else:
            if self.on_generate:
                self.on_generate(self)
        if self.on_after_generate:
            self.on_after_generate(self)
        return self.report_url

    def delete_report(self, file_name):
        report_name = os.path.join(self.task.work_dir, 'static', 'reports', file_name)
        os.remove(report_name)

    def parse_template(self):
        if not os.path.isabs(self.template):
            self.template_name = os.path.join(self.task.work_dir, 'reports', self.template)
        else:
            self.template_name = self.template
        z = zipfile.ZipFile(self.template_name, 'r')
        try:
            data = z.read('content.xml')
        finally:
            z.close()
        band_tags = []
        bands = {}
        colum_defs = []
        header = ''
        columns = ''
        rows = ''
        footer = ''
        repeated_rows = None
        if data:
            dom = parseString(data)
            try:
                tables = dom.getElementsByTagName('table:table')
                if len(tables) > 0:
                    table = tables[0]
                    for child in table.childNodes:
                        if child.nodeName == 'table:table-column':
                            repeated = child.getAttribute('table:number-columns-repeated')
                            if not repeated:
                                repeated = 1
                            colum_defs.append(['', repeated])
                        if child.nodeName == 'table:table-row':
                            repeated = child.getAttribute('table:number-rows-repeated')
                            if repeated and repeated.isdigit():
                                repeated_rows = str(repeated)
                            for row_child in child.childNodes:
                                if row_child.nodeName == 'table:table-cell':
                                    text = row_child.getElementsByTagName('text:p')
                                    if text.length > 0:
                                        band_tags.append(text[0].childNodes[0].nodeValue)
                                    break
                start = 0
                columns_start = 0
                for col in colum_defs:
                    start = data.find('<table:table-column', start)
                    if columns_start == 0:
                        columns_start = start
                    end = data.find('/>', start)
                    col_text = data[start: end + 2]
                    columns += col_text
                    col[0] = data[start: end + 2]
                    start = end + 2
                columns_end = start
                header = data[0:columns_start]
                assert len(band_tags) > 0, u'No bands in report template'
                positions = []
                start = 0
                for tag in band_tags:
                    text = str('>%s<' % tag)
                    i = data.find(text)
                    i = data.rfind('<table:table-row', start, i)
                    positions.append(i)
                    start = i
                if repeated_rows and int(repeated_rows) > 1000:
                    i = data.find(repeated_rows)
                    i = data.rfind('<table:table-row', start, i)
                    band_tags.append('$$$end_of_report')
                    positions.append(i)
                rows = data[columns_end:positions[0]]
                for i, tag in enumerate(band_tags):
                    start = positions[i]
                    try:
                        end = positions[i + 1]
                    except:
                        end = data.find('</table:table>', start)
                    bands[tag] = data[start: end].replace(str(tag), '')
                footer = data[end:len(data)]
                self.template_content = {}
                self.template_content['bands'] = bands
                self.template_content['colum_defs'] = colum_defs
                self.template_content['header'] = header
                self.template_content['columns'] = columns
                self.template_content['rows'] = rows
                self.template_content['footer'] = footer
            finally:
                dom.unlink()
                del(dom)

    def hide_columns(self, col_list):

        def convert_str_to_int(string):
            s = string.upper()
            base = ord('A')
            mult = ord('Z') - base + 1
            result = s
            if type(s) == str:
                result = 0
                chars = []
                for i in range(len(s)):
                    chars.append(s[i])
                for i in range(len(chars) - 1, -1, -1):
                    result += (ord(chars[i]) - base + 1) * (mult ** (len(chars) - i - 1))
            return result

        def remove_repeated(col, repeated):
            result = col
            p = col.find('table:number-columns-repeated')
            if p != -1:
                r = col.find(str(repeated), p)
                if r != -1:
                    for i in xrange(r, 100):
                        if col[i] in ("'", '"'):
                            result = col.replace(col[p:i+1], '')
                            break
            return result

        if self.template_content:
            ints = []
            for i in col_list:
                ints.append(convert_str_to_int(i))
            colum_defs = self.template_content['colum_defs']
            columns = ''
            index = 1
            for col, repeated in colum_defs:
                repeated = int(repeated)
                if repeated > 1:
                    col = remove_repeated(col, repeated)
                for i in range(repeated):
                    cur_col = col
                    if index in ints:
                        cur_col = cur_col[0:-2] + ' table:visibility="collapse"/>'
                    columns += cur_col
                    index += 1
            self.template_content['colum_defs'] = colum_defs
            self.template_content['columns'] = columns

    def print_band(self, band, dic=None, update_band_text=None):
        text = self.template_content['bands'][band]
        if dic:
            d = dic.copy()
            for key, value in d.items():
                if type(value) in (str, unicode):
                    d[key] = escape(value)
            cell_start = 0
            cell_start_tag = '<table:table-cell'
            cell_type_tag = 'office:value-type="string"'
            calcext_type_tag = 'calcext:value-type="string"'
            start_tag = '<text:p>'
            end_tag = '</text:p>'
            while True:
                cell_start = text.find(cell_start_tag, cell_start)
                if cell_start == -1:
                    break
                else:
                    start = text.find(start_tag, cell_start)
                    if start != -1:
                        end = text.find(end_tag, start + len(start_tag))
                        if end != -1:
                            text_start = start+len(start_tag)
                            text_end = end
                            cell_text = text[text_start:text_end]
                            cell_text_start = cell_text.find('%(', 0)
                            if cell_text_start != -1:
                                end = cell_text.find(')s', cell_text_start + 2)
                                if end != -1:
                                    end += 2
                                    val = cell_text[cell_text_start:end]
                                    key = val[2:-2]
                                    value = d.get(key)
                                    if not value is None:
                                        val = val % d
                                        if type(value) == float:
                                            val = val.replace('.', common.DECIMAL_POINT)
                                    else:
                                        if not key in d.keys():
                                            print 'Report: "%s" band: "%s" key "%s" not found in the dictionary' % (self.item_name, band, key)
                                    if  isinstance(cell_text, unicode):
                                        cell_text = cell_text.encode('utf8')
                                    if  isinstance(val, unicode):
                                        val = val.encode('utf8')
                                    cell_text = cell_text[:cell_text_start] + val + cell_text[end:]
                                    text = text[:text_start] + cell_text + text[text_end:]
                                    if type(value) in (int, float):
                                        start_text = text[cell_start:start]
                                        office_value = str(value)
                                        start_text = start_text.replace(cell_type_tag, 'office:value-type="float" office:value="%s"' % office_value)
                                        start_text = start_text.replace(calcext_type_tag, 'calcext:value-type="float"')
                                        text = text[:cell_start] + start_text + text[start:]
                    cell_start += 1
            if update_band_text:
                text = update_band_text(text)
        self.content.write(text)

    def save(self):
        self.content.close()
        z = None
        self.zip_file = None
        try:
            self.zip_file = zipfile.ZipFile(self.report_filename, 'w', zipfile.ZIP_DEFLATED)
            z = zipfile.ZipFile(self.template_name, 'r')
            if self.on_before_save_report:
                self.on_before_save_report(self)
            for file_name in z.namelist():
                data = z.read(file_name)
                if file_name == 'content.xml':
                    self.zip_file.write(self.content_name, file_name)
                else:
                    self.zip_file.writestr(file_name, data)
        finally:
            if z:
                z.close()
            if self.zip_file:
                self.zip_file.close()

    def cur_to_str(self, value):
        return common.cur_to_str(value)

    def date_to_str(self, value):
        return common.date_to_str(value)

    def datetime_to_str(self, value):
        return common.datetime_to_str(value)

    def _set_modified(self, value):
        pass

#~ log = {}
#~ log_counter = 0
#~ log_save_after = 0
#~ log_pid = os.getpid()
#~
#~ def find_where_fields(command):
    #~ result = ''
    #~ where_pos = command.find('WHERE')
    #~ where = None
    #~ if where_pos >= 0:
        #~ a = []
        #~ where = command[where_pos + 5:]
        #~ a.append(where.find(' GROUP BY'))
        #~ a.append(where.find(' ORDER BY'))
        #~ a.append(where.find(' HAVING'))
        #~ pos = -1
        #~ b = []
        #~ for p in a:
            #~ if p >= 0:
                #~ b.append(p)
        #~ if len(b):
            #~ pos = min(b)
        #~ if pos >= 0:
            #~ where = where[:pos + 1]
        #~ if where:
            #~ parts = where.split('AND')
            #~ fields = []
            #~ for p in parts:
                #~ p = p.strip()
                #~ p = p.strip('(')
                #~ w = ''
                #~ for ch in p:
                    #~ if ch in [' ', '>', '<', '=']:
                        #~ break
                    #~ else:
                        #~ w += ch
                #~ w = w.replace('"', '')
                #~ pos = w.find('.')
                #~ if pos >= 0:
                    #~ w = w[pos + 1:]
                #~ fields.append(w)
            #~ result = ' '.join(fields)
    #~ return result
#~
#~ def find_table_name(command):
    #~ result = ''
    #~ s = ''
    #~ if command[:12] ==   'SELECT NEXT ':
        #~ result = ''
    #~ elif command[:12] == 'INSERT INTO ':
        #~ result = ''
    #~ elif command[:6] == 'UPDATE':
        #~ s = command[7:]
        #~ s = s.lstrip('"')
    #~ else:
        #~ pos = command.find(' FROM ')
        #~ s = command[pos+5:pos+105]
        #~ s = s.lstrip(' ')
        #~ s = s.lstrip('(')
        #~ s = s.lstrip('"')
    #~ if s:
        #~ w = ''
        #~ for ch in s:
            #~ if ch in [' ', '"']:
                #~ break
            #~ else:
                #~ w += ch
        #~ result = w
    #~ return result
#~
#~ def save_log(log):
#~
    #~ def get_reqs_key(item):
        #~ key, (dur, freq) = item
        #~ return dur / 1000000.0 / freq
#~
    #~ def get_tables_key(item):
        #~ key, (dur, freq, reqs) = item
        #~ return dur / 1000000.0 / freq
#~
    #~ text = ''
#~
    #~ tlist = sorted(log.iteritems(), key=get_tables_key, reverse=True)
    #~ for table, (dur, freq, reqs) in tlist:
        #~ sec = dur / 1000000.0
        #~ per_request = sec / freq
        #~ text += '%-20s %f %d %f\n' % (table, per_request, freq, sec)
#~
    #~ text += '\n'
#~
    #~ tables = sorted(log.keys())
    #~ for table in tables:
        #~ dur, freq, reqs = log[table]
        #~ sec = dur / 1000000.0
        #~ per_request = sec / freq
        #~ text += '\n%-20s %f %d %f\n' % (table, per_request, freq, sec)
        #~ rlist = sorted(reqs.iteritems(), key=get_reqs_key, reverse=True)
        #~ for fields, (dur, freq) in rlist:
            #~ sec = dur / 1000000.0
            #~ per_request = sec / freq
            #~ text += '      %f %d %f %s\n' % (per_request, freq, sec, fields)
    #~ with open('log.info', 'w') as f:
        #~ f.write(text.encode('utf-8'))
#~
#~
#~ def log_request(command, duration):
    #~ global log_counter
    #~ global log_pid
    #~ global log
    #~ global log_save_after
#~
    #~ if log_pid == os.getpid():
        #~ command = command.upper().strip()
        #~ table_name = find_table_name(command)
        #~ if table_name:
            #~ where_fields = find_where_fields(command)
#~
            #~ msec = duration.microseconds
#~
            #~ table = log.get(table_name)
            #~ if not table:
                #~ table = [0, 0, {}]
                #~ log[table_name] = table
            #~ table[0] += msec
            #~ table[1] += 1
            #~ reqs = table[2]
#~
            #~ req = reqs.get(where_fields)
            #~ if not req:
                #~ req = [0, 0]
                #~ reqs[where_fields] = req
            #~ req[0] += msec
            #~ req[1] += 1
#~
        #~ log_counter += 1
        #~ if log_counter % log_save_after == 0:
            #~ save_log(log)
#~

def execute_sql(db_module, db_database, db_user, db_password,
    db_host, db_port, db_encoding, connection, command,
    params=None, call_proc=False, commit=True, select=False, ddl=False):

    #~ def execute_command(cursor, command, params=None):
        #~ global log_save_after
        #~ print_command = False
        #~ log_command = False
        #~ try:
            #~ if print_command:
                #~ print ''
                #~ print command
                #~ print params
#~
            #~ now = datetime.datetime.now()
#~
            #~ result = None
            #~ if params:
                #~ cursor.execute(command, params)
            #~ else:
                #~ cursor.execute(command)
            #~ if result_set == 'ONE':
                #~ result = cursor.fetchone()
            #~ elif result_set == 'ALL':
                #~ result = cursor.fetchall()
#~
            #~ if log_save_after:
                #~ log_request(command, datetime.datetime.now() - now)
#~
            #~ if log_command and command.upper().find('SELECT') == -1:
                #~ try:
                    #~ with open("sql_log.txt", "a") as f:
                        #~ f.write('\n')
                        #~ f.write(command + '\n')
                        #~ f.write(json.dumps(params, default=common.json_defaul_handler) + '\n')
                #~ except:
                    #~ pass
#~
            #~ return result
        #~ except Exception, x:
            #~ error = '\nError: %s\n command: %s\n params: %s' % (str(x), command, params)
            #~ print error
            #~ if not ddl:
                #~ raise
            #~ elif db_module.DDL_ROLLBACK:
                #~ raise

    def execute_command(cursor, command, params=None):
        print_command = ddl
        try:
            if print_command:
                print ''
                print command
                if params:
                    print params
                    messages.append('<p>' + command + '<br>' + \
                        json.dumps(params, default=common.json_defaul_handler) + '</p>')
                else:
                    messages.append('<p>' + command + '</p>')

            now = datetime.datetime.now()

            result = None
            if params and not select:
                cursor.execute(command, db_module.process_sql_params(params, cursor))
            else:
                cursor.execute(command)
            if select:
                result = cursor.fetchall()
                result = db_module.process_sql_result(result)
            return result
        except Exception, x:
            error = '\nError: %s\n command: %s\n params: %s' % (str(x), command, params)
            print error
            if ddl:
                arr = str(x).split('\\n')
                error = '<br>'.join(arr)
                messages.append('<div class="text-error">%s</div>' % error)
                if db_module.DDL_ROLLBACK:
                    raise
            else:
                raise

    def get_next_id(cursor, sql):
        cursor.execute(sql)
        rec = cursor.fetchone()
        return int(rec[0])

    def execute_delta(cursor, command, delta_result):

        def process_delta(delta, master_rec_id, result):
            ID, sqls = delta
            result['ID'] = ID
            changes = []
            result['changes'] = changes
            for sql in sqls:
                (command, params, info, h_sql, h_params, h_table_name), details = sql
                if info:
                    rec_id = info['primary_key']
                    if info['status'] == common.RECORD_INSERTED:
                        if rec_id:
                            pass
                        else:
                            next_sequence_value_sql = db_module.next_sequence_value_sql(info['table_name'])
                            if next_sequence_value_sql:
                                rec_id = get_next_id(cursor, next_sequence_value_sql)
                                params[info['primary_key_index']] = rec_id
                    if info['status'] == common.RECORD_INSERTED and info['master_rec_id_index']:
                        params[info['master_rec_id_index']] = master_rec_id
                    if command:
                        execute_command(cursor, command, params)
                    if not rec_id and info['status'] == common.RECORD_INSERTED:
                        new_id = db_module.get_lastrowid(cursor)
                        if new_id:
                            rec_id = new_id
                    result_details = []
                    if rec_id:
                        changes.append({'log_id': info['log_id'], 'rec_id': rec_id, 'details': result_details})
                    for detail in details:
                        result_detail = {}
                        result_details.append(result_detail)
                        process_delta(detail, rec_id, result_detail)
                else:
                    if command:
                        execute_command(cursor, command, params)
                if h_sql:
                    next_sequence_value_sql = db_module.next_sequence_value_sql(h_table_name)
                    if next_sequence_value_sql:
                        h_id = get_next_id(cursor, next_sequence_value_sql)
                        h_params[0] = h_id
                    if not h_params[2]:
                        h_params[2] = rec_id
                    execute_command(cursor, h_sql, h_params)

        delta = command['delta']
        process_delta(delta, None, delta_result)

    def execute_list(cursor, command, delta_result):
        res = None
        if command:
            for com in command:
                if com:
                    if isinstance(com, unicode) or isinstance(com, str):
                        res = execute_command(cursor, com)
                    elif isinstance(com, list):
                        res = execute_list(cursor, com, delta_result)
                    elif isinstance(com, dict):
                        res = execute_delta(cursor, com, delta_result)
                    elif isinstance(com, tuple):
                        res = execute_command(cursor, com[0], com[1])
                    else:
                        raise Exception, 'server_classes execute_list: invalid argument - command: %s' % command
            return res

    def execute(connection):
        delta_result = {}
        result = None
        error = None
        try:
            cursor = connection.cursor()
            if call_proc:
                try:
                    cursor.callproc(command, params)
                    result = cursor.fetchone()
                except Exception, x:
                    print '\nError: %s in command: %s' % (str(x), command)
                    raise
            else:
                if isinstance(command, str) or isinstance(command, unicode):
                    result = execute_command(cursor, command, params)
                elif isinstance(command, dict):
                    res = execute_delta(cursor, command, delta_result)
                elif isinstance(command, list):
                    result = execute_list(cursor, command, delta_result)
                elif isinstance(command, tuple):
                    result = execute_command(cursor, command[0], command[1])
            if commit:
                connection.commit()
            else:
                connection.rollback()
            if delta_result:
                result = delta_result
        except Exception, x:
            try:
                if connection:
                    connection.rollback()
                    connection.close()
                error = str(x)
                if not error:
                    error = 'SQL execution error'
                print traceback.format_exc()
            finally:
                connection = None
        finally:
            if select:
                pass
            if ddl:
                info = ''
                if messages:
                    info = "".join(messages)
                return connection, (result, error, info)
            else:
                return connection, (result, error)

    messages = []
    if connection is None:
        try:
            connection = db_module.connect(db_database, db_user, db_password, db_host, db_port, db_encoding)
        except Exception, x:
             print str(x)
             return  None, (None, str(x))
    return execute(connection)

def process_request(parentPID, name, queue, db_type, db_database, db_user, db_password, db_host, db_port, db_encoding, mod_count):
    con = None
    counter = 0
    db_module = db_modules.get_db_module(db_type)
    while True:
        if parentPID and hasattr(os, 'getppid') and os.getppid() != parentPID:
            break
        request = queue.get()
        if request:
            result_queue = request['queue']
            command = request['command']
            params = request['params']
            call_proc = request['call_proc']
            commit = request['commit']
            select = request['select']
            cur_mod_count = request['mod_count']
            if cur_mod_count != mod_count or counter > 1000:
                if con:
                    con.rollback()
                    con.close()
                con = None
                mod_count = cur_mod_count
                counter = 0
            con, result = execute_sql(db_module, db_database, db_user, db_password,
                db_host, db_port, db_encoding, con, command, params, call_proc, commit, select)
            counter += 1
            result_queue.put(result)

class Consts(object):
    def __init__(self):
        self.TEXT = common.TEXT
        self.INTEGER = common.INTEGER
        self.FLOAT = common.FLOAT
        self.CURRENCY = common.CURRENCY
        self.DATE = common.DATE
        self.DATETIME = common.DATETIME
        self.BOOLEAN = common.BOOLEAN
        self.BLOB = common.BLOB

        self.ITEM_FIELD = common.ITEM_FIELD
        self.FILTER_FIELD = common.FILTER_FIELD
        self.PARAM_FIELD = common.PARAM_FIELD

        self.FILTER_EQ = common.FILTER_EQ
        self.FILTER_NE = common.FILTER_NE
        self.FILTER_LT = common.FILTER_LT
        self.FILTER_LE = common.FILTER_LE
        self.FILTER_GT = common.FILTER_GT
        self.FILTER_GE = common.FILTER_GE
        self.FILTER_IN = common.FILTER_IN
        self.FILTER_NOT_IN = common.FILTER_NOT_IN
        self.FILTER_RANGE = common.FILTER_RANGE
        self.FILTER_ISNULL = common.FILTER_ISNULL
        self.FILTER_EXACT = common.FILTER_EXACT
        self.FILTER_CONTAINS = common.FILTER_CONTAINS
        self.FILTER_STARTWITH = common.FILTER_STARTWITH
        self.FILTER_ENDWITH = common.FILTER_ENDWITH
        self.FILTER_CONTAINS_ALL = common.FILTER_CONTAINS_ALL

        self.ALIGN_LEFT = common.ALIGN_LEFT
        self.ALIGN_CENTER = common.ALIGN_CENTER
        self.ALIGN_RIGHT = common.ALIGN_RIGHT

        self.STATE_INACTIVE = common.STATE_INACTIVE
        self.STATE_BROWSE = common.STATE_BROWSE
        self.STATE_INSERT = common.STATE_INSERT
        self.STATE_EDIT = common.STATE_EDIT
        self.STATE_DELETE = common.STATE_DELETE

        self.RECORD_UNCHANGED = common.RECORD_UNCHANGED
        self.RECORD_INSERTED = common.RECORD_INSERTED
        self.RECORD_MODIFIED = common.RECORD_MODIFIED
        self.RECORD_DETAILS_MODIFIED = common.RECORD_DETAILS_MODIFIED
        self.RECORD_DELETED = common.RECORD_DELETED


class AbstractServerTask(AbstrTask):
    def __init__(self, app, name, caption, js_filename, db_type,
        db_database = '', db_user = '', db_password = '', host='', port='',
        encoding='', con_pool_size=1, mp_pool=False, persist_con=False):
        AbstrTask.__init__(self, None, None, None, None)
        self.app = app
        self.consts = Consts()
        self.items = []
        self.lookup_lists = {}
        self.ID = None
        self.item_name = name
        self.item_caption = caption
        self.js_filename = js_filename
        self.db_type = db_type
        self.db_database = db_database
        self.db_user = db_user
        self.db_password = db_password
        self.db_host = host
        self.db_port = port
        self.db_encoding = encoding
        self.db_module = db_modules.get_db_module(self.db_type)
        self.on_request = None
        self.work_dir = os.getcwd()
        self.con_pool_size = 0
        self.mod_count = 0
        self.modules = []
        self.conversion_lock = threading.Lock()
        self.con_pool_size = con_pool_size
        self.mp_pool = mp_pool
        self.persist_con = persist_con
        self.persist_con_busy = 0
        if self.mp_pool:
            if self.persist_con:
                self.create_connection_pool(1)
            self.create_mp_connection_pool(self.con_pool_size)
        else:
            self.create_connection_pool(self.con_pool_size)

    def get_version(self):
        return common.SETTINGS['VERSION']

    version = property (get_version)

    def create_connection_pool(self, con_count):
        self.queue = Queue.Queue()
        pid = None
        for i in range(con_count):
            p = threading.Thread(target=process_request, args=(pid, self.item_name,
                self.queue, self.db_type, self.db_database, self.db_user,
                self.db_password, self.db_host, self.db_port,
                self.db_encoding, self.mod_count))
            p.daemon = True
            p.start()

    def create_mp_connection_pool(self, con_count):
        self.mp_queue = multiprocessing.Queue()
        self.mp_manager = multiprocessing.Manager()
        pid = os.getpid()
        for i in range(con_count):
            p = multiprocessing.Process(target=process_request, args=(pid, self.item_name,
                self.mp_queue, self.db_type, self.db_database, self.db_user,
                self.db_password, self.db_host, self.db_port,
                self.db_encoding, self.mod_count))
            p.daemon = True
            p.start()

    def create_connection(self):
        return self.db_module.connect(self.db_database, self.db_user, self.db_password, self.db_host, self.db_port, self.db_encoding)

    def send_to_pool(self, queue, result_queue, command, params=None, call_proc=False, commit=True, select=False):
        request = {}
        request['queue'] = result_queue
        request['command'] = command
        request['params'] = params
        request['call_proc'] = call_proc
        request['commit'] = commit
        request['select'] = select
        request['mod_count'] = self.mod_count
        queue.put(request)
        return  result_queue.get()

    def execute_in_pool(self, command, params=None, call_proc=False, commit=True, select=False):
        result_queue = Queue.Queue()
        result = self.send_to_pool(self.queue, result_queue, command, params, call_proc, commit, select)
        return result

    def execute_in_mp_poll(self, command, params=None, call_proc=False, commit=True, select=False):
        result_queue = self.mp_manager.Queue()
        result = self.send_to_pool(self.mp_queue, result_queue, command, params, call_proc, commit, select)
        return result

    def execute(self, command, params=None, call_proc=False, commit=True, select=False):
        if self.mp_pool:
            if self.persist_con and not self.persist_con_busy:
                self.persist_con_busy += 1
                try:
                    result = self.execute_in_pool(command, params, call_proc, commit, select)
                finally:
                    self.persist_con_busy -= 1
            else:
                result = self.execute_in_mp_poll(command, params, call_proc, commit, select)
        else:
            result = self.execute_in_pool(command, params, call_proc, commit, select)
        return result

    def callproc(self, command, params=None):
        result_set, error = self.execute(command, params, call_proc=True)
        if not error:
            return result_set

    def execute_select(self, command, params=None):
        result, error = self.execute(command, params, commit=False, select=True)
        if error:
            raise Exception, error
        else:
            return result

    def get_module_name(self):
        return str(self.item_name)

    def compile_item(self, item):
        item.module_name = None
        code = item.server_code
        item.module_name = item.get_module_name()
        item_module = type(sys)(item.module_name)
        item_module.__dict__['this'] = item
        sys.modules[item.module_name] = item_module

        item.task.modules.append(item.module_name)
        if item.owner:
            sys.modules[item.owner.get_module_name()].__dict__[item.module_name] = item_module
        if code:
            try:
                code = code.encode('utf-8')
            except Exception, e:
                print e
            comp_code = compile(code, item.module_name, "exec")
            exec comp_code in item_module.__dict__

            item_module.__dict__['__loader__'] = item._loader
            funcs = inspect.getmembers(item_module, inspect.isfunction)
            item._events = []
            for func_name, func in funcs:
                item._events.append((func_name, func))
                setattr(item, func_name, func)
        del code

    def add_item(self, item):
        self.items.append(item)
        item.owner = self
        return item

    def find_item(self, g_index, i_index):
        return self.items[g_index].items[i_index]

    def convert_report(self, report, ext):
        converted = False
        with self.conversion_lock:
            try:
                from subprocess import Popen, STDOUT, PIPE
                if os.name == "nt":
                    import _winreg
                    regpath = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\soffice.exe"
                    root = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, regpath)
                    s_office = _winreg.QueryValue(root, "")
                else:
                    s_office = "soffice"
                convertion = Popen([s_office, '--headless', '--convert-to', ext,
                    report.report_filename, '--outdir', os.path.join(self.work_dir, 'static', 'reports') ],
                    stderr=STDOUT,stdout = PIPE)#, shell=True)
                out, err = convertion.communicate()
                converted = True
            except Exception, e:
                print e
        return converted

    def copy_database(self, dbtype, database=None, user=None, password=None,
        host=None, port=None, encoding=None):
        connection = None
        limit = 1024
        db_module = db_modules.get_db_module(dbtype)
        for group in self.items:
            for it in group.items:
                if it.item_type != 'report':
                    item = it.copy(handlers=False, filters=False, details=False)
                    if item.table_name and not item.virtual_table:
                        self.execute(self.db_module.set_case('DELETE FROM %s' % item.table_name))
                        item.open(expanded=False, open_empty=True)
                        params = {'__fields': [], '__filters': [], '__expanded': False, '__offset': 0, '__limit': 0}
                        sql = item.get_record_count_query(params, db_module)
                        connection, (result, error) = \
                        execute_sql(db_module, database, user, password,
                            host, port, encoding, connection, sql, None, 'ALL')
                        record_count = result[0][0]
                        loaded = 0
                        max_id = 0
                        if record_count:
                            while True:
                                params['__offset'] = loaded
                                params['__limit'] = limit
                                sql = item.get_select_statement(params, db_module)
                                connection, (result, error) = \
                                execute_sql(db_module, database, user, password,
                                    host, port, encoding, connection, sql, None, 'ALL')
                                if not error:
                                    for i, r in enumerate(result):
                                        item.append()
                                        j = 0
                                        for field in item.fields:
                                            if not field.master_field:
                                                field.value = r[j]
                                                j += 1
                                        if item._primary_key_field.value > max_id:
                                            max_id = item._primary_key_field.value
                                        item.post()
                                    item.apply()
                                else:
                                    raise Exception, error
                                records = len(result)
                                loaded += records
                                print 'coping table %s: %d%%' % (item.item_name, int(loaded * 100 / record_count))
                                if records == 0 or records < limit:
                                    break
                            if self.db_module.restart_sequence_sql:
                                sql = self.db_module.restart_sequence_sql(item.table_name, max_id + 1)
                                self.execute(sql)

class DebugException(Exception):
    pass


class Task(AbstractServerTask):
    def __init__(self, app, name, caption, js_filename,
        db_type, db_database = '', db_user = '', db_password = '',
        host='', port='', encoding='', con_pool_size=4, mp_pool=True,
        persist_con=True):
        AbstractServerTask.__init__(self, app, name, caption, js_filename,
            db_type, db_database, db_user, db_password,
            host, port, encoding, con_pool_size, mp_pool, persist_con)
        self.on_created = None
        self.on_login = None
        self.on_get_user_info = None
        self.on_logout = None
        self.on_ext_request = None
        self.init_dict = {}
        for key, value in self.__dict__.items():
            self.init_dict[key] = value

class AdminTask(AbstractServerTask):
    def __init__(self, app, name, caption, js_filename,
        db_type, db_database = '', db_user = '', db_password = '',
        host='', port='', encoding=''):
        AbstractServerTask.__init__(self, app, name, caption, js_filename,
            db_type, db_database, db_user, db_password, host, port, encoding, 2)
        filepath, filename = os.path.split(__file__)
        self.cur_path = filepath
        self.users = {}
        self.roles = None
        self.edited_docs = []

    def create_task(self):
        return adm_server.create_task(self.app)

    def reload_task(self):
        adm_server.reload_task(self)

    def update_events_code(self):
        adm_server.update_events_code(self)

    def login(self, log, psw_hash, is_admin, env):
        privileges = None
        if not is_admin and self.app.task and self.app.task.on_login:
            user_info = self.app.task.on_login(self.app.task, log, psw_hash, env)
        else:
            user_info = adm_server.login(self, log, psw_hash, is_admin)
        user_uuid = None
        if user_info:
            for key in self.users.iterkeys():
                if self.users[key]["user_id"] == user_info["user_id"] and \
                    self.users[key]["admin"] == user_info["admin"]:
                    if user_info["admin"]:
                        adm_server.clear_docs_on_logout(self, user_info)
                    del self.users[key]
                    break
            user_uuid = str(uuid.uuid4())
            self.users[user_uuid] = user_info
        return user_uuid, ''

    def logout(self, user_uuid, admin, env):
        try:
            user_info = self.users.get(user_uuid)
            adm_server.clear_docs_on_logout(self, user_info)
            if user_info:
                if not admin and self.app.task and self.app.task.on_logout:
                    self.app.task.on_logout(self.app.task, user_info, env)
                del self.users[user_uuid]
        except:
            pass
        return None, ''

    def get_user_info(self, user_uuid, admin, env):
        user_info = self.users.get(user_uuid)
        if user_info:
            if not admin or (admin and user_info['admin']):
                return user_info

    def get_privileges(self, role_id):
        if self.roles is None:
            self.roles = adm_server.get_roles(self)
        return self.roles[role_id]

    def find_privileges(self, user_info, item):
        if not self.safe_mode or item.master or (item.task == self) or (item == item.task):
            return {'can_view': True, 'can_create': True, 'can_edit': True, 'can_delete': True}
        else:
            try:
                priv_dic = self.get_privileges(user_info['role_id'])[item.ID]
            except:
                priv_dic = None
            if priv_dic:
                return priv_dic
            else:
                return {'can_view': False, 'can_create': False, 'can_edit': False, 'can_delete': False}

    def has_privilege(self, user_info, item, priv_name):
        return self.find_privileges(user_info, item)[priv_name]

class Group(AbstrGroup):
    def __init__(self, owner, name, caption, view_template=None, js_filename=None, visible=True, item_type_id=0):
        AbstrGroup.__init__(self, owner, name, caption, visible, item_type_id, js_filename)
        self.ID = None
        self.view_template = view_template
        self.js_filename = js_filename
        if item_type_id == common.REPORTS_TYPE:
            self.on_convert_report = None

    def add_catalog(self, name, caption, table_name, visible=True, view_template='', js_filename='', soft_delete=True):
        result = Item(self, name, caption, visible, table_name, view_template, js_filename, soft_delete)
        result.item_type_id = common.ITEM_TYPE
        return result

    def add_table(self, name, caption, table_name, visible=True, view_template='', js_filename='', soft_delete=True):
        result = Item(self, name, caption, visible, table_name, view_template, js_filename, soft_delete)
        result.item_type_id = common.TABLE_TYPE
        return result

    def add_report(self, name, caption, table_name, visible=True, view_template='', js_filename='', soft_delete=True):
        result = Report(self, name, caption, visible, table_name, view_template, js_filename)
        result.item_type_id = common.REPORT_TYPE
        return result


class Detail(AbstrDetail, ServerDataset):
    def __init__(self, owner, name, caption, table_name):
        AbstrDetail.__init__(self, owner, name, caption, True)
        ServerDataset.__init__(self, table_name)
        self.prototype = self.task.item_by_name(self.item_name)
        self.master = owner

    def init_fields(self):
        self.field_defs = []
        for field_def in self.prototype.field_defs:
            self.field_defs.append(list(field_def))
        for field_def in self.field_defs:
            field = DBField(self, field_def)
            self._fields.append(field)
        self._primary_key = self.prototype._primary_key
        self._deleted_flag = self.prototype._deleted_flag
        self._master_id = self.prototype._master_id
        self._master_rec_id = self.prototype._master_rec_id


    def do_internal_post(self):
        return {'success': True, 'id': None, 'message': '', 'detail_ids': None}

    def where_clause(self, query, db_module):
        master_id = query['__master_id']
        master_rec_id = query['__master_rec_id']
        if type(master_id) == int and type(master_rec_id) == int:
            result = super(Detail, self).where_clause(query, db_module)
            clause = '%s."%s"=%s AND %s."%s"=%s' % \
                (self.table_alias(), self._master_id, str(master_id),
                self.table_alias(), self._master_rec_id, str(master_rec_id))
            if result:
                result += ' AND ' + clause
            else:
                result = ' WHERE ' + clause
            return db_module.set_case(result)
        else:
            raise Exception, 'Invalid request parameter'

    def get_filters(self):
        return self.prototype.filters

    def get_reports_info(self):
        return []



# -*- coding: utf-8 -*-

SQLITE, FIREBIRD, POSTGRESQL, MYSQL, ORACLE = range(1, 6)
DB_TYPE = ('Sqlite', 'FireBird', 'PostgreSQL', 'MySQL', 'Oracle')

def get_db_module(db_type):
    db = None
    if db_type == SQLITE:
        import sqlite as db
    elif db_type == FIREBIRD:
        import firebird as db
    elif db_type == POSTGRESQL:
        import postgres as db
    elif db_type == MYSQL:
        import mysql as db
    elif db_type == ORACLE:
        import oracle as db
    return db



# -*- coding: utf-8 -*-

import MySQLdb

NEED_DATABASE_NAME = True
NEED_LOGIN = True
NEED_PASSWORD = True
NEED_ENCODING = True
NEED_HOST = True
NEED_PORT = True
CAN_CHANGE_TYPE = False
CAN_CHANGE_SIZE = False
UPPER_CASE = True
DDL_ROLLBACK = False
FROM = '"%s" AS %s'
LEFT_OUTER_JOIN = 'LEFT OUTER JOIN "%s" AS %s'
FIELD_AS = 'AS'
LIKE = 'LIKE'

JAM_TYPES = TEXT, INTEGER, FLOAT, CURRENCY, DATE, DATETIME, BOOLEAN, BLOB = range(1, 9)
FIELD_TYPES = {
    INTEGER: 'INT',
    TEXT: 'VARCHAR',
    FLOAT: 'DOUBLE',
    CURRENCY: 'DOUBLE',
    DATE: 'DATE',
    DATETIME: 'DATETIME',
    BOOLEAN: 'INT',
    BLOB: 'BLOB'
}

def connect(database, user, password, host, port, encoding):
    charset = None
    use_unicode = None
    if encoding:
        charset = encoding
        use_unicode = True
    if port:
        connection = MySQLdb.connect(db=database, user=user, passwd=password, host=host,
            port=int(port), charset=charset, use_unicode=use_unicode)
    else:
        connection = MySQLdb.connect(db=database, user=user, passwd=password, host=host,
            charset=charset, use_unicode=use_unicode)
    connection.autocommit(False)
    cursor = connection.cursor()
    cursor.execute("SET SESSION SQL_MODE=ANSI_QUOTES")
    return connection

def get_lastrowid(cursor):
    return cursor.lastrowid

def get_select(query, start, end, fields):
    offset = query['__offset']
    limit = query['__limit']
    result = 'SELECT %s FROM %s' % (start, end)
    if limit:
        result += ' LIMIT %d, %d' % (offset, limit)
    return result

def process_sql_params(params, cursor):
    result = []
    for p in params:
        if type(p) == tuple:
            value, data_type = p
        else:
            value = p
        result.append(value)
    return result

def process_sql_result(rows):
    result = []
    for row in rows:
        result.append(list(row))
    return result

def cast_date(date_str):
    return "CAST('" + date_str + "' AS DATE)"

def cast_datetime(datetime_str):
    return "CAST('" + datetime_str + "' AS TIMESTAMP)"

def value_literal(index):
    return '%s'

def upper_function():
    pass

def create_table_sql(table_name, fields, foreign_fields=None):
    result = []
    primary_key = ''
    sql = set_case('CREATE TABLE "%s"\n(\n' % table_name)
    for field in fields:
        sql += set_case('"%s" %s' % (field['field_name'], FIELD_TYPES[field['data_type']]))
        if field['size'] != 0 and field['data_type'] == TEXT:
            sql += '(%d)' % field['size']
        if field['default_value'] and not field['primary_key']:
            if field['data_type'] == TEXT:
                sql += " DEFAULT '%s'" % field['default_value']
            else:
                sql += ' DEFAULT %s' % field['default_value']
        if field['primary_key']:
            sql += ' NOT NULL AUTO_INCREMENT'
            primary_key = set_case(field['field_name'])
        sql +=  ',\n'
    sql += 'PRIMARY KEY("%s")' % primary_key
    sql += ')\n'
    result.append(sql)
    return result

def delete_table_sql(table_name):
    result = []
    result.append(set_case('DROP TABLE "%s"' % table_name))
    return result

def create_index_sql(index_name, table_name, unique, fields, desc):
    return set_case('CREATE %s INDEX "%s" ON "%s" (%s)' % \
        (unique, index_name, table_name, fields))

def create_foreign_index_sql(table_name, index_name, key, ref, primary_key):
    return set_case('ALTER TABLE %s ADD CONSTRAINT %s FOREIGN KEY (%s) REFERENCES %s(%s)' % \
        (table_name, index_name, key, ref, primary_key))

def delete_index(table_name, index_name):
    return 'DROP INDEX %s ON %s' % (index_name, table_name)

def delete_foreign_index(table_name, index_name):
    return set_case('ALTER TABLE %s DROP FOREIGN KEY %s' % \
        (table_name, index_name))

def add_field_sql(table_name, field):
    result = set_case('ALTER TABLE "%s" ADD "%s" %s' % \
        (table_name, field['field_name'], FIELD_TYPES[field['data_type']]))
    if field['size']:
        result += '(%d)' % field['size']
    if field['default_value']:
        if field['data_type'] == TEXT:
            result += " DEFAULT '%s'" % field['default_value']
        else:
            result += ' DEFAULT %s' % field['default_value']
    return result

def del_field_sql(table_name, field):
    return set_case('ALTER TABLE "%s" DROP "%s"' % (table_name, field['field_name']))

def change_field_sql(table_name, old_field, new_field):
    result = []
    if old_field['field_name'] != new_field['field_name']:
        sql = set_case('ALTER TABLE "%s" CHANGE  "%s" "%s" %s' % (table_name, old_field['field_name'],
            new_field['field_name'], FIELD_TYPES[new_field['data_type']]))
        if old_field['size']:
            sql += '(%d)' % old_field['size']
        #~ if old_field['data_type'] != new_field['data_type'] or \
            #~ old_field['size'] != new_field['size']:
            #~ sql += set_case(' %s' % FIELD_TYPES[field['data_type']])
            #~ if new_field['size'] and old_field['size'] != new_field['size']:
                #~ sql += '(%d)' % new_field['size']
        result.append(sql)
    if old_field['default_value'] != new_field['default_value']:
        if new_field['default_value']:
            if new_field['data_type'] == TEXT:
                sql = set_case('ALTER TABLE "%s" ALTER "%s" SET DEFAULT' % \
                    (table_name, new_field['field_name']))
                sql +=  " '%s'" % new_field['default_value']
            else:
                sql = set_case('ALTER TABLE "%s" ALTER "%s" SET DEFAULT %s' % \
                    (table_name, new_field['field_name'], new_field['default_value']))
        else:
            sql = set_case('ALTER TABLE "%s" ALTER "%s" DROP DEFAULT' % \
                (table_name, new_field['field_name']))
        result.append(sql)
    return result

def set_case(string):
    return string.upper()

def param_literal():
    return '%s'

def quotes():
    return '`'

def get_sequence_name(table_name):
    return None

def next_sequence_value_sql(table_name):
    return None

def restart_sequence_sql(table_name, value):
    pass



# -*- coding: utf-8 -*-

import datetime
import traceback
import common

FIELD_DEF = FIELD_ID, FIELD_NAME, NAME, FIELD_DATA_TYPE, REQUIRED, LOOKUP_ITEM, MASTER_FIELD, LOOKUP_FIELD, \
    FIELD_VISIBLE, FIELD_VIEW_INDEX, FIELD_EDIT_VISIBLE, FIELD_EDIT_INDEX, FIELD_READ_ONLY, FIELD_EXPAND, FIELD_WORD_WRAP, \
    FIELD_SIZE, FIELD_DEFAULT_VALUE, FIELD_DEFAULT, FIELD_CALCULATED, FIELD_EDITABLE, FIELD_ALIGNMENT, FIELD_LOOKUP_VALUES, \
    FIELD_ENABLE_TYPEAHEAD, FIELD_HELP, FIELD_PLACEHOLDER = range(25)

FILTER_DEF = FILTER_OBJ_NAME, FILTER_NAME, FILTER_FIELD_NAME, FILTER_TYPE, FILTER_DATA_TYPE, FILTER_VISIBLE, \
    FILTER_HELP, FILTER_PLACEHOLDER = range(8)

class DatasetException(Exception):
    pass

class DBField(object):
    def __init__(self, owner, field_def):
        self.owner = owner
        self.field_def = field_def
        self.field_kind = common.ITEM_FIELD
        self.ID = field_def[FIELD_ID]
        self.field_name = field_def[FIELD_NAME]
        self.field_caption = field_def[NAME]
        self.data_type = field_def[FIELD_DATA_TYPE]
        self.required = field_def[REQUIRED]
        self.lookup_item = field_def[LOOKUP_ITEM]
        self.master_field = field_def[MASTER_FIELD]
        self.lookup_field = field_def[LOOKUP_FIELD]
        self.read_only = field_def[FIELD_READ_ONLY]
        self.view_visible = field_def[FIELD_VISIBLE]
        self.view_index = field_def[FIELD_VIEW_INDEX]
        self.edit_visible = field_def[FIELD_EDIT_VISIBLE]
        self.edit_index = field_def[FIELD_EDIT_INDEX]
        self.expand = field_def[FIELD_EXPAND]
        self.word_wrap = field_def[FIELD_WORD_WRAP]
        self.field_size = field_def[FIELD_SIZE]
        self.default_value = field_def[FIELD_DEFAULT_VALUE]
        self.is_default = field_def[FIELD_DEFAULT]
        self.calculated = field_def[FIELD_CALCULATED]
        self.editable = field_def[FIELD_EDITABLE]
        self.alignment = field_def[FIELD_ALIGNMENT]
        self.lookup_values = field_def[FIELD_LOOKUP_VALUES]
        self.field_type = common.FIELD_TYPE_NAMES[self.data_type]
        self.filter = None
        self.on_get_field_text_called = None

    def get_row(self):
        if self.owner._dataset:
            return self.owner._dataset[self.owner.rec_no]
        else:
            print traceback.format_exc()
            raise Exception, u'An attempt to get a field value in the empty dataset - item: %s, field: %s ' % (self.owner.item_name, self.field_name)

    def get_data(self):
        row = self.get_row()
        if row and self.bind_index >= 0:
            return row[self.bind_index]

    def set_data(self, value):
        row = self.get_row()
        if row and (self.bind_index >= 0):
            row[self.bind_index] = value

    def get_lookup_data(self):
        if self.lookup_index:
            row = self.get_row()
            if row and (self.lookup_index >= 0):
                return row[self.lookup_index]

    def set_lookup_data(self, value):
        if self.lookup_index:
            row = self.get_row()
            if row and (self.lookup_index >= 0):
                row[self.lookup_index] = value

    def update_controls(self):
        pass

    def get_text(self):
        result = ''
        try:
            result = self.get_value()
            if not result is None:
                if self.data_type == common.INTEGER:
                    result = str(result)
                elif self.data_type == common.FLOAT:
                    result = self.float_to_str(result)
                elif self.data_type == common.CURRENCY:
                    result = self.cur_to_str(result)
                elif self.data_type == common.DATE:
                    result = self.date_to_str(result)
                elif self.data_type == common.DATETIME:
                    result = self.datetime_to_str(result)
                elif self.data_type == common.TEXT:
                    result = unicode(result)
                elif self.data_type == common.BOOLEAN:
                    if self.value:
                        if self.owner:
                            result = self.owner.task.lang['yes']
                        else:
                            result = u'True'
                    else:
                        if self.owner:
                            result = self.owner.task.lang['no']
                        else:
                            result = u'False'
                else:
                    result = str(result)
            else:
                result = ''
        except Exception, e:
            print traceback.format_exc()
            self.do_on_error(self.type_error() % (''), e)
        return result

    def set_text(self, value):
        if value != self.text:
            try:
                if self.data_type == common.TEXT:
                    self.set_value(unicode(value))
                elif self.data_type == common.INTEGER:
                    self.set_value(int(value))
                if self.data_type == common.FLOAT:
                    self.set_value(common.str_to_float(value))
                elif self.data_type == common.CURRENCY:
                    self.set_value(common.str_to_float(value))
                elif self.data_type == common.DATE:
                    self.set_value(common.str_to_date(value))
                elif self.data_type == common.DATETIME:
                    self.set_value(common.str_to_datetime(value))
                elif self.data_type == common.BOOLEAN:
                    if self.owner:
                        if len(value) and value.upper().split() == self.owner.task.lang['yes'].upper().split():
                            self.set_value(True)
                        else:
                            self.set_value(False)
                    else:
                        if len(value) and value[0] in [u'T', u't']:
                            self.set_value(True)
                        else:
                            self.set_value(False)
                else:
                    self.set_value(value)
            except Exception, e:
                print traceback.format_exc()
                self.do_on_error(self.type_error() % (value), e)

    text = property (get_text, set_text)

    def convert_date_time(self, value):
        if value.find('.'):
            value = value.split('.')[0]
        if value.find('T'):
            value = value.replace('T', ' ')
        return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')

    def convert_date(self, value):
        try:
            if value.find(' '):
                value = value.split(' ')[0]
            return datetime.datetime.strptime(value, '%Y-%m-%d').date()
        except:
            return self.convert_date_time(value).date()


    def get_raw_value(self):
        try:
            value = self.get_data()
            if self.data_type == common.DATE:
                if type(value) in [str, unicode]:
                    value = self.convert_date(value)
            elif self.data_type == common.DATETIME:
                if type(value) in [str, unicode]:
                    value = self.convert_date_time(value)
            return value
        except Exception, e:
            print traceback.format_exc()
            self.do_on_error(self.type_error() % (''), e)

    raw_value = property (get_raw_value)

    def get_value(self):
        value = self.get_raw_value()
        try:
            if value == None:
                if self.field_kind == common.ITEM_FIELD:
                    if self.data_type in (common.FLOAT, common.INTEGER, common.CURRENCY):
                        value = 0
                    elif self.data_type == common.BOOLEAN:
                        value = False
                    elif self.data_type == common.TEXT:
                        value = ''
            else:
                if self.data_type in (common.TEXT, ):
                    if not isinstance(value, unicode):
                        value = value.decode('utf-8')
                elif self.data_type == common.BOOLEAN:
                    if value:
                        value = True
                    else:
                        value = False
            return value
        except Exception, e:
            print traceback.format_exc()
            self.do_on_error(self.type_error() % (''), e)

    def _change_lookup_field(self, lookup_value=None):
        if self.lookup_item:
            if self.owner:
                master_field = self
                if self.master_field:
                    master_field = self.master_field
                master_field.lookup_value = None
                for field in self.owner.fields:
                    if field.master_field == master_field:
                        field.lookup_value = None
            if lookup_value:
                self.lookup_value = lookup_value

    def _do_before_changed(self):
        if self.owner:
            if not self.owner.item_state in (common.STATE_INSERT, common.STATE_EDIT):
                raise DatasetException, u'%s is not in edit or insert mode' % self.owner.item_name
            if self.owner.on_before_field_changed:
                self.owner.on_before_field_changed(self)

    def _check_system_field_value(self, value):
        if self.field_kind == common.ITEM_FIELD:
            if self.field_name == self.owner._primary_key and self.value and self.value != value:
                raise DatasetException, u'%s: сhanging of the primary field is forbidden' % self.owner.item_name
            if self.field_name == self.owner._deleted_flag and self.value != value:
                raise DatasetException, u'%s, field - %s: сhanging of the deleted flag field is forbidden' % (self.owner.item_name, self.field_name)

    def _is_filter_value(self, value):
        if self.field_kind == common.FILTER_FIELD:
            if self.filter.filter_type in (common.FILTER_IN, common.FILTER_NOT_IN):
                if not type(value) in [list, tuple]:
                    error = u'Filter "%s" value must be an array' % self.filter.filter_name
                    self.do_on_error(error);
                return True

    def set_value(self, value, lookup_value=None, lookup_item=None):
        self._check_system_field_value(value)
        self.new_value = None
        if not value is None:
            self.new_value = value
            if not self._is_filter_value(value):
                if self.data_type == common.BOOLEAN:
                    if bool(value):
                        self.new_value = 1
                    else:
                        self.new_value = 0
                elif self.data_type in (common.FLOAT, common.CURRENCY):
                    self.new_value = float(value)
                elif self.data_type == common.INTEGER:
                    self.new_value = int(value)
        if self.raw_value != self.new_value:
            self._do_before_changed()
            try:
                self.set_data(self.new_value)
            except Exception, e:
                print traceback.format_exc()
                self.do_on_error(self.type_error() % (value), e)
            finally:
                self.new_value = None
            self._change_lookup_field(lookup_value)
            self._set_modified(True)
            self._do_after_changed(lookup_item)

    value = property (get_value, set_value)

    def _set_modified(self, value):
        if not self.calculated:
            if self.owner:
                self.owner._set_modified(value)

    def get_lookup_data_type(self):
        if self.lookup_item:
            return self.lookup_item._field_by_name(self.lookup_field).data_type;

    def get_lookup_value(self):
        value = None
        if self.lookup_item:
            if self.value:
                value = self.get_lookup_data()
                data_type = self.get_lookup_data_type()
                if data_type == common.DATE:
                    if isinstance(value, unicode):
                        value = self.convert_date(value)
                elif data_type == common.DATETIME:
                    if isinstance(value, unicode):
                        value = self.convert_date_time(value)
                elif self.data_type == common.BOOLEAN:
                    value = bool(value)
        return value

    def set_lookup_value(self, value):
        if self.lookup_item:
            self.set_lookup_data(value)

    lookup_value = property (get_lookup_value, set_lookup_value)

    def get_lookup_text(self):
        result = ''
        try:
            if self.lookup_item:
                if self.value:
                    result = self.get_lookup_value()
                if result is None:
                    result = ''
                else:
                    data_type = self.get_lookup_data_type()
                    if data_type:
                        if data_type == common.DATE:
                            result = self.date_to_str(result)
                        elif data_type == common.DATETIME:
                            result = self.datetime_to_str(result)
                        elif data_type == common.FLOAT:
                            result = self.float_to_str(result)
                        elif data_type == common.CURRENCY:
                            result = self.cur_to_str(result)
            result = unicode(result)
        except Exception, e:
            print traceback.format_exc()
        return result

    lookup_text = property (get_lookup_text)

    def get_display_text(self):

        def get_value_in_list():
            result = '';
            for val, str_val in self.lookup_values:
                if val == self.value:
                    result = str_val
            return result

        result = ''
        if self.filter and self.filter.filter_type == common.FILTER_IN and self.filter.field.lookup_item and self.filter.value:
            result = self.filter.owner.task.lang['items_selected'] % len(self.filter.value)
        elif self.lookup_item:
            result = self.lookup_text
        elif self.lookup_values and self.value:
            try:
                result = get_value_in_list()
            except:
                pass
        else:
            if self.data_type == common.CURRENCY:
                if not self.raw_value is None:
                    result = common.currency_to_str(self.value)
            else:
                result = self.text
        if self.owner and not self.filter:
            if self.owner.on_get_field_text:
                if not self.on_get_field_text_called:
                    self.on_get_field_text_called = True
                    try:
                        res = self.owner.on_get_field_text(self)
                        if not res is None:
                            result = res
                    finally:
                        self.on_get_field_text_called = False
        return result

    display_text = property(get_display_text)

    def _set_read_only(self, value):
        self._read_only = value
        self.update_controls()

    def _get_read_only(self):
        result = self._read_only
        if self.owner and self.owner.parent_read_only and self.owner.read_only:
            result = self.owner.read_only
        return result

    read_only = property (_get_read_only, _set_read_only)

    def set_visible(self, value):
        self._visible = value
        self.update_controls()

    def get_visible(self):
        return self._visible
    view_visible = property (get_visible, set_visible)

    def set_alignment(self, value):
        self._alignment = value
        self.update_controls()

    def get_alignment(self):
        return self._alignment
    alignment = property (get_alignment, set_alignment)

    def xalign(self):
        return (self.alignment - 1) * 0.5

    def set_expand(self, value):
        self._expand = value
        self.update_controls()

    def get_expand(self):
        return self._expand
    expand = property (get_expand, set_expand)

    def set_word_wrap(self, value):
        self._word_wrap = value
        self.update_controls()

    def get_word_wrap(self):
        return self._word_wrap
    word_wrap = property (get_word_wrap, set_word_wrap)

    def check_type(self):
        self.get_value()
        if (self.data_type == common.TEXT) and (self.field_size != 0) and \
            (len(self.text) > self.field_size):
            print self.text, len(self.text), type(self.text)
            self.do_on_error(self.owner.task.lang['invalid_length'] % self.field_size)
        return True

    def check_reqired(self):
        if not self.required:
            return True
        elif self.get_raw_value() != None:
            return True
        else:
            self.do_on_error(self.owner.task.lang['value_required'])

    def check_valid(self):
        if self.check_reqired():
            if self.check_type():
                if self.owner:
                    if self.owner.on_field_validate:
                        e = self.owner.on_field_validate(self)
                        if e:
                            self.do_on_error(e)
                return True

    def _do_after_changed(self, lookup_item):
        if self.owner:
            if self.owner.on_field_changed:
                self.owner.on_field_changed(self, lookup_item)
        self.update_controls()

    def do_on_error(self, err_mess, err=None):
        mess = None
        if err_mess and not err:
            mess = err_mess
        elif isinstance(err, ValueError) and err_mess:
            mess = err_mess
        elif err:
            mess = str(err)
        if mess:
            if self.owner:
                print u'%s %s - %s: %s' % (self.owner.item_name,
                    self.owner.task.lang['error'].lower(), self.field_name, mess)
        raise TypeError(mess)

    def type_error(self):
        if self.data_type == common.INTEGER:
            return self.owner.task.lang['invalid_int']
        elif self.data_type == common.FLOAT:
            return self.owner.task.lang['invalid_float']
        elif self.data_type == common.CURRENCY:
            return self.owner.task.lang['invalid_cur']
        elif (self.data_type == common.DATE) or (self.data_type == common.DATE):
            return self.owner.task.lang['invalid_date']
        elif self.data_type == common.BOOLEAN:
            return self.owner.task.lang['invalid_bool']
        else:
            return self.owner.task.lang['invalid_value']

    def system_field(self):
        if self.field_name and self.field_name in (self.owner._primary_key,  \
            self.owner._deleted_flag, self.owner._master_id, self.owner._master_rec_id):
            return True

    def float_to_str(self, value):
        return common.float_to_str(value)

    def cur_to_str(self, value):
        return common.cur_to_str(value)

    def date_to_str(self, value):
        return common.date_to_str(value)

    def datetime_to_str(self, value):
        return common.datetime_to_str(value)

    def str_to_date(self, value):
        return common.str_to_date(value)

    def str_to_datetime(self, value):
        return common.str_to_datetime(value)

    def str_to_float(self, value):
        return common.str_to_float(value)

    def str_to_cur(self, value):
        return common.str_to_currency(value)

class FilterField(DBField):
    def __init__(self, fltr, field, owner):
        DBField.__init__(self, owner, field.field_def)
        self.field_kind = common.FILTER_FIELD
        self.filter = fltr
        self.lookup_item = None
        self._value = None
        self._lookup_value = None


    def _do_before_changed(self):
        pass

    def get_data(self):
        return self._value

    def set_data(self, value):
        self._value = value

    def get_lookup_data(self):
        return self._lookup_value

    def set_lookup_data(self, value):
        self._lookup_value = value

    def check_reqired(self, value):
        return True

    def _set_modified(self, value):
        pass

    def _set_record_status(self, value):
        pass

class DBFilter(object):
    def __init__(self, owner, filter_def):
        self.owner = owner
        self.filter_def = filter_def
        self.filter_name = filter_def[FILTER_OBJ_NAME]
        self.filter_caption = filter_def[FILTER_NAME]
        self.field_name = filter_def[FILTER_FIELD_NAME]
        self.filter_type = filter_def[FILTER_TYPE]
        self.data_type = filter_def[FILTER_DATA_TYPE]
        self.visible = filter_def[FILTER_VISIBLE]
        self.filter_help = filter_def[FILTER_HELP]
        if type(self.field_name) == int:
            self.field_name = owner._field_by_ID(self.field_name).field_name
        field = self.owner._field_by_name(self.field_name)
        if self.field_name:
            self.field = FilterField(self, field, self.owner)
            setattr(self, self.field_name, self.field)
            if self.filter_type == common.FILTER_RANGE:
                self.field1 = FilterField(self, field, self.owner)

    def set_value(self, value):
        if self.filter_type == common.FILTER_RANGE:
            if value is None:
                self.field.value = None;
                self.field1.value = None;
            else:
                self.field.value = value[0];
                self.field1.value = value[1];
        else:
            self.field.value = value

    def get_value(self):
        return self.field.raw_value

    value = property (get_value, set_value)

class DBList(list):
    pass

class ChangeLog(object):
    def __init__(self, item):
        self.item = item
        self._change_id = 0
        self.records = []
        self.logs = {}
        self.fields = []
        self.expanded = True

    def get_change_id(self):
        self._change_id += 1
        return str(self._change_id)

    def log_changes(self):
        if self.item.master:
            return self.item.master.change_log.log_changes()
        else:
            return self.item.log_changes;

    def find_record_log(self):
        result = None
        if self.item.master:
            record_log = self.item.master.change_log.find_record_log()
            if record_log:
                details = record_log['details']
                detail = details.get(str(self.item.ID))
                if not detail:
                    detail = {
                        'logs': {},
                        'records': self.item._dataset,
                        'fields': [field.field_name for field in self.item.fields],
                        'expanded': self.item.expanded
                    }
                    details[str(self.item.ID)] = detail
                self.logs = detail['logs']
                self.records = detail['records']
                self.fields = detail['fields']
                self.expanded = detail['expanded']
        if self.item.record_count():
            change_id = self.item.rec_change_id
            if not change_id:
                change_id = self.get_change_id()
                self.item.rec_change_id = change_id;
            result = self.logs.get(change_id)
            if not result:
                result = {
                    'old_record': None,
                    'record': self.cur_record(),
                    'details': {}
                }
                self.logs[change_id] = result
            return result

    def get_detail_log(self, detail_ID):
        result = None
        record_log = self.find_record_log()
        details = record_log['details']
        if details:
            result = details.get(detail_ID)
        if result is None and self.item._is_delta:
            result = {'records': [], 'fields': [], 'expanded': False, 'logs': {}}
        return result

    def remove_record_log(self):
        change_id = self.item.rec_change_id
        if change_id:
            self.find_record_log()
            del self.logs[self.item.rec_change_id]
            self.item.rec_change_id = None
            self.item.record_status = common.RECORD_UNCHANGED

    def cur_record(self):
        return self.item._dataset[self.item.rec_no]

    def record_modified(self, record_log):
        modified = False
        old_rec = record_log['old_record']
        cur_rec = record_log['record']
        for i in xrange(self.item._record_lookup_index):
            if old_rec[i] != cur_rec[i]:
                modified = True
                break
        return modified

    def copy_record(self, record, expanded=True):
        result = None
        if record:
            if expanded:
                result = record[0:self.item._record_info_index]
            else:
                result = record[0:self.item._record_lookup_index]
            info = self.item.get_rec_info(record=record)
            result.append([info[0], {}, info[2]])
        return result

    def log_change(self):
        if self.log_changes():
            record_log = self.find_record_log()
            if self.item.item_state == common.STATE_BROWSE:
                if (self.item.record_status == common.RECORD_UNCHANGED) or \
                    (self.item.record_status == common.RECORD_DETAILS_MODIFIED and record_log['old_record'] is None):
                    record_log['old_record'] = self.copy_record(self.cur_record(), False)
                    return
            elif self.item.item_state == common.STATE_INSERT:
                self.item.record_status = common.RECORD_INSERTED
            elif self.item.item_state == common.STATE_EDIT:
                if self.item.record_status == common.RECORD_UNCHANGED:
                    self.item.record_status = common.RECORD_MODIFIED
                elif self.item.record_status == common.RECORD_DETAILS_MODIFIED:
                    if self.record_modified(record_log):
                        self.item.record_status = common.RECORD_MODIFIED
            elif self.item.item_state == common.STATE_DELETE:
                if self.item.record_status == common.RECORD_INSERTED:
                    self.remove_record_log()
                else:
                    self.item.record_status = common.RECORD_DELETED
            else:
                raise Exception, u'%s: change log invalid records state' % self.item.item_name
            if self.item.master:
                if self.item.master.record_status == common.RECORD_UNCHANGED:
                    self.item.master.record_status = common.RECORD_DETAILS_MODIFIED

    def get_changes(self, result):
        data = {}
        result['fields'] = self.fields
        result['expanded'] = False
        result['data'] = data
        for key, record_log in self.logs.iteritems():
            record = record_log['record']
            info = self.item.get_rec_info(record=record)
            if info[common.REC_STATUS] != common.RECORD_UNCHANGED:
                old_record = record_log['old_record']
                new_record = self.copy_record(record_log['record'], expanded=False)
                new_details = {}
                for detail_id, detail in record_log['details'].iteritems():
                    new_detail = {}
                    detail_item = self.item.item_by_ID(int(detail_id))
                    detail_item.change_log.logs = detail['logs']
                    detail_item.change_log.fields = detail['fields']
                    detail_item.change_log.expanded = detail['expanded']
                    detail_item.change_log.get_changes(new_detail)
                    new_details[detail_id] = new_detail
                data[key] = {
                        'old_record': old_record,
                        'record': new_record,
                        'details': new_details
                    }

    def set_changes(self, changes):
        new_records = []
        self.records = []
        self.logs = {}
        self.fields = changes['fields']
        self.expanded = changes['expanded']
        data = changes['data']
        self._change_id = 0
        for key, record_log in data.iteritems():
            if self._change_id < int(key):
                self._change_id = int(key)
            record = record_log['record']
            record[len(record) - 1].append(record_log['old_record'])
            new_records.append([int(key), record])
            details = {}
            self.logs[key] = {
                'old_record': None,
                'record': record,
                'details': details
            }
            for detail_id, detail in record_log['details'].iteritems():
                detail_item = self.item.item_by_ID(int(detail_id))
                detail_item.change_log.set_changes(detail)
                details[detail_id] = {
                    'logs': detail_item.change_log.logs,
                    'records': detail_item.change_log.records,
                    'fields': detail_item.change_log.fields,
                    'expanded': detail_item.change_log.expanded
                }
        new_records.sort(key=lambda x: x[0])
        self.records = [rec for key, rec in new_records]

    def copy_records(self, records):
        result = []
        for rec in records:
            result.append(list(rec))
        return result

    def store_details(self, source, dest):
        for detail_item in self.item.details:
            detail_id = str(detail_item.ID)
            detail = source.get(detail_id)
            logs = {}
            records = []
            fields = []
            expanded = True
            if detail:
                cur_logs = detail['logs']
                cur_records = detail['records']
                fields = detail['fields']
                expanded = detail['expanded']
                records = self.copy_records(cur_records)
                for key, record_log in cur_logs.iteritems():
                    cur_record = record_log['record']
                    record = detail_item.change_log.copy_record(cur_record)
                    index = None
                    try:
                        index = cur_records.index(cur_record)
                    except:
                        pass
                    if not index is None:
                        records[index] = record
                    details = {}
                    detail_item.change_log.store_details(record_log['details'], details)
                    logs[key] = {
                        'old_record': record_log['old_record'],
                        'record': record,
                        'details': details
                    }
            else:
                if detail_item._dataset:
                    records = self.copy_records(detail_item._dataset)
            if records or logs:
                dest[detail_id] = {'logs': logs, 'records': records, 'fields': fields, 'expanded': expanded}

    def store_record_log(self):
        if not self.log_changes():
            result = {}
            result['record'] = self.copy_record(self.cur_record())
            details = {}
            for detail in self.item.details:
                if not detail.disabled and detail._dataset:
                    details[str(detail.ID)] = list(detail._dataset)
            result['details'] = details
        else:
            record_log = self.find_record_log()
            details = {}
            self.store_details(record_log['details'], details)
            result = {}
            result['old_record'] = record_log['old_record']
            result['record'] = self.copy_record(record_log['record'])
            result['details'] = details
        return result

    def restore_record_log(self, log):

        def restore_record():
            record = log['record']
            cur_record = self.cur_record()
            info_index = self.item._record_info_index
            for i, it in enumerate(cur_record):
                if i < info_index:
                    cur_record[i] = record[i]

        if not self.log_changes():
            restore_record()
            for detail in self.item.details:
                detail._dataset = log['details'][str(detail.ID)]
        else:
            record_log = self.find_record_log()
            restore_record()
            record_log['old_record'] = log['old_record']
            record_log['record'] = self.cur_record()
            record_log['details'] = log['details']
            for detail in self.item.details:
                detail_log = log['details'].get(str(detail.ID))
                if detail_log:
                    detail._dataset = detail_log['records']
            if self.item.record_status == common.RECORD_UNCHANGED:
                self.remove_record_log()


    def prepare(self):
        if self.item.master:
            log = self.item.master.change_log.get_detail_log(str(self.item.ID))
            if log:
                log['records'] = []
                log['logs'] = {}
                log['fields'] = []
                log['fields'] = [field.field_name for field in self.item.fields if not field.master_field]
                log['expanded'] = self.item.expanded
        else:
            self.records = []
            self.logs = {}
            self.fields = []
            self.fields = [field.field_name for field in self.item.fields if not field.master_field]
            self.expanded = self.item.expanded

    def update(self, updates, master_rec_id=None):
        if updates:
            changes = updates['changes']
            for change in changes:
                log_id = change['log_id']
                rec_id = change['rec_id']
                details = change['details']
                record_log = self.logs.get(log_id)
                if record_log:
                    record = record_log['record']
                    record_details = record_log['details']
                    for detail in details:
                        ID = detail['ID']
                        detail_item = self.item.detail_by_ID(int(ID))
                        item_detail = record_details.get(str(ID))
                        if item_detail:
                            detail_item.change_log.logs = item_detail['logs']
                            detail_item.change_log.update(detail, rec_id)
                    if rec_id:
                        if not record[self.item._primary_key_field.bind_index]:
                            record[self.item._primary_key_field.bind_index] = rec_id
                    if master_rec_id:
                        if not record[self.item._master_rec_id_field.bind_index]:
                            record[self.item._master_rec_id_field.bind_index] = master_rec_id
                    info = self.item.get_rec_info(record=record)
                    info[common.REC_STATUS] = common.RECORD_UNCHANGED
                    info[common.REC_CHANGE_ID] = common.RECORD_UNCHANGED
                    del self.logs[log_id]


class AbstractDataSet(object):
    def __init__(self):
        self.ID = 0
        self.field_defs = []
        self._fields = []
        self.fields = []
        self.filter_defs = []
        self.filters = DBList()
        self.details = DBList()
        self.controls = []
        self.change_log = ChangeLog(self)
        self._log_changes = True
        self._dataset = None
        self._primary_key = None
        self._deleted_flag = None
        self._master_id = None
        self._master_rec_id = None
        self._eof = False
        self._bof = False
        self._cur_row = None
        self._old_row = 0
        self._old_status = None
        self._buffer = None
        self._modified = None
        self._state = common.STATE_INACTIVE
        self._read_only = False
        self._active = False
        self._where_list = []
        self._order_by_list = []
        self._select_field_list = []
        self.on_state_changed = None
        self.on_filter_changed = None
        self._record_lookup_index = -1
        self._record_info_index = -1
        self._filtered = False
        self.expanded = True
        self._open_params = {}
        self.post_local = False
        self._disabled_count = 0
        self.open_params = {}
        self._is_delta = False
        self.keep_history = False
        self.parent_read_only = True
        self.on_before_append = None
        self.on_after_append = None
        self.on_before_edit = None
        self.on_after_edit = None
        self.on_before_open = None
        self.on_after_open = None
        self.on_before_post = None
        self.on_after_post = None
        self.on_before_delete = None
        self.on_after_delete = None
        self.on_before_cancel = None
        self.on_after_cancel = None
        self.on_before_apply = None
        self.on_after_apply = None
        self.on_before_scroll = None
        self.on_after_scroll = None
        self.on_filter_record = None
        self.on_field_changed = None
        self.on_filters_applied = None
        self.on_before_field_changed = None
        self.on_filter_value_changed = None
        self.on_field_validate = None
        self.on_get_field_text = None

    def __getitem__(self, key):
        if key == 0:
            self.first()
        else:
            self.next()
        if self.eof():
            raise IndexError
        return self

    def add_field_def(self, field_ID, field_name, field_caption, data_type, required, lookup_item, lookup_field,
            view_visible, view_index, edit_visible, edit_index, read_only, expand, word_wrap,
            field_size, default_value, is_default, calculated, editable, master_field, alignment,
            lookup_values, enable_typeahead, field_help, field_placeholder):
        field_def = [None for i in range(len(FIELD_DEF))]
        field_def[FIELD_ID] = field_ID
        field_def[FIELD_NAME] = field_name
        field_def[NAME] = field_caption
        field_def[FIELD_DATA_TYPE] = data_type
        field_def[REQUIRED] = required
        field_def[LOOKUP_ITEM] = lookup_item
        field_def[MASTER_FIELD] = master_field
        field_def[LOOKUP_FIELD] = lookup_field
        field_def[FIELD_READ_ONLY] = read_only
        field_def[FIELD_VISIBLE] = view_visible
        field_def[FIELD_VIEW_INDEX] = view_index
        field_def[FIELD_EDIT_VISIBLE] = edit_visible
        field_def[FIELD_EDIT_INDEX] = edit_index
        field_def[FIELD_EXPAND] = expand
        field_def[FIELD_WORD_WRAP] = word_wrap
        field_def[FIELD_SIZE] = field_size
        field_def[FIELD_DEFAULT_VALUE] = default_value
        field_def[FIELD_DEFAULT] = is_default
        field_def[FIELD_CALCULATED] = calculated
        field_def[FIELD_EDITABLE] = editable
        field_def[FIELD_ALIGNMENT] = alignment
        field_def[FIELD_LOOKUP_VALUES] = lookup_values
        field_def[FIELD_ENABLE_TYPEAHEAD] = enable_typeahead
        field_def[FIELD_HELP] = field_help
        field_def[FIELD_PLACEHOLDER] = field_placeholder
        self.field_defs.append(field_def)
        return field_def

    def add_filter_def(self, filter_name, filter_caption, field_name, filter_type, \
        data_type, visible, filter_help, filter_placeholder):
        filter_def = [None for i in range(len(FILTER_DEF))]
        filter_def[FILTER_OBJ_NAME] = filter_name
        filter_def[FILTER_NAME] = filter_caption
        filter_def[FILTER_FIELD_NAME] = field_name
        filter_def[FILTER_TYPE] = filter_type
        filter_def[FILTER_DATA_TYPE] = data_type
        filter_def[FILTER_VISIBLE] = visible
        filter_def[FILTER_HELP] = filter_help
        filter_def[FILTER_PLACEHOLDER] = filter_placeholder
        self.filter_defs.append(filter_def)
        return filter_def

    def get_dataset(self):
        result = []
        if self.active:
            for r in self._dataset:
                result.append(r[0:self._record_info_index])
        return result

    def set_dataset(self, value):
        self._dataset = value

    dataset = property (get_dataset, set_dataset)

    def _copy(self, filters=True, details=True, handlers=True):
        result = self.__class__(self.owner, self.item_name, self.item_caption, self.visible)
        result.ID = self.ID
        result.item_name = self.item_name
        result.expanded = self.expanded
        result.field_defs = self.field_defs
        result.filter_defs = self.filter_defs
        result.keep_history = self.keep_history

        for field_def in result.field_defs:
            field = DBField(result, field_def)
            result._fields.append(field)
        result.prepare_fields()

        for filter_def in result.filter_defs:
            fltr = DBFilter(result, filter_def)
            result.filters.append(fltr)
        result.prepare_filters()

        result._events = self._events
        if handlers:
            for func_name, func in result._events:
                setattr(result, func_name, func)
        return result

    def clone(self, keep_filtered=True):
        result = self.__class__(self.owner, self.item_name, self.item_caption, self.visible)
        result.ID = self.ID
        result.item_name = self.item_name
        result.field_defs = self.field_defs
        result.filter_defs = self.filter_defs

        for field_def in result.field_defs:
            field = DBField(result, field_def)
            result._fields.append(field)
        result.prepare_fields()

        for field in result.fields:
            if hasattr(result, field.field_name):
                delattr(result, field.field_name)

        result.fields = [];
        for field in self.fields:
            new_field = result._field_by_name(field.field_name)
            result.fields.append(new_field)
            if not hasattr(result, new_field.field_name):
                setattr(result, new_field.field_name, new_field)
        result._bind_fields()
        result._dataset = self._dataset
        if keep_filtered:
            result.on_filter_record = self.on_filter_record
            result.filtered = self.filtered
        result._active = True
        result.first()
        return result

    def prepare_fields(self):

        for field in self._fields:
            if field.lookup_item and type(field.lookup_item) == int:
                field.lookup_item = self.task.item_by_ID(field.lookup_item)
            if field.master_field and type(field.master_field) == int:
                field.master_field = self._field_by_ID(field.master_field)
            if field.lookup_field and type(field.lookup_field) == int:
                field.lookup_field = field.lookup_item._field_by_ID(field.lookup_field).field_name
            if field.lookup_values and type(field.lookup_values) == int:
                try:
                    field.lookup_values = self.task.lookup_lists[field.lookup_values]
                except:
                    pass
        self.fields = list(self._fields)
        for field in self.fields:
            if not hasattr(self, field.field_name):
                setattr(self, field.field_name, field)
        for sys_field_name in ['_primary_key', '_deleted_flag', '_master_id', '_master_rec_id']:
            sys_field = getattr(self, sys_field_name)
            if sys_field and type(sys_field) == int:
                field = self.field_by_ID(sys_field)
                if field:
                    setattr(self, sys_field_name, field.field_name)

    def prepare_filters(self):
        for fltr in self.filters:
            setattr(self.filters, fltr.filter_name, fltr)
            if fltr.field.lookup_item and type(fltr.field.lookup_item) == int:
                fltr.field.lookup_item = self.task.item_by_ID(fltr.field.lookup_item)

    def field_by_name(self, field_name):
        for field in self.fields:
            if field.field_name == field_name:
                return field

    def _field_by_name(self, field_name):
        for field in self._fields:
            if field.field_name == field_name:
                return field

    def field_by_ID(self, id_value):
        for field in self.fields:
            if field.ID == id_value:
                return field

    def _field_by_ID(self, id_value):
        for field in self._fields:
            if field.ID == id_value:
                return field

    def filter_by_name(self, name):
        for fltr in self.filters:
            if fltr.filter_name == name:
                return fltr

    def _get_log_changes(self):
        return self._log_changes

    def _set_log_changes(self, value):
        self._log_changes = value

    log_changes = property (_get_log_changes, _set_log_changes)

    def _set_modified(self, value):
        self._modified = value

    def is_modified(self):
        return self._modified

    def disable_controls(self):
        self._disabled_count -= 1

    def enable_controls(self):
        self._disabled_count += 1
        if self.controls_enabled():
            self.update_controls(common.UPDATE_SCROLLED)

    def controls_enabled(self):
        return self._disabled_count == 0

    def controls_disabled(self):
        return not self.controls_enabled()

    def _get_active(self):
        return self._active

    active = property (_get_active)

    def _set_read_only(self, value):
        self._read_only = value
        for field in self.fields:
            field.update_controls()

    def _get_read_only(self):
        return self._read_only

    read_only = property (_get_read_only, _set_read_only)

    def _get_filtered(self):
        return self._filtered

    def _set_filtered(self, value):
        if value:
            if not self.on_filter_record:
                value = False
        if self._filtered != value:
            self._filtered = value
            self.first()
            self.update_controls(common.UPDATE_OPEN)

    filtered = property (_get_filtered, _set_filtered)

    def _set_item_state(self, value):
        if self._state != value:
            self._state = value
            if self.on_state_changed:
                self.on_state_changed(self)

    def _get_item_state(self):
        return self._state

    item_state = property (_get_item_state, _set_item_state)

    def _do_after_scroll(self):
        self.update_controls(common.UPDATE_SCROLLED)
        if self.on_after_scroll:
            self.on_after_scroll(self)

    def _do_before_scroll(self):
        if not self._cur_row is None:
            if self.item_state in (common.STATE_INSERT, common.STATE_EDIT):
                self.post()
            if self.on_before_scroll:
                self.on_before_scroll(self)

    def skip(self, value):
        if self.record_count() == 0:
            self._do_before_scroll()
            self._eof = True
            self._bof = True
            self._do_after_scroll()
        else:
            old_row = self._cur_row
            eof = False
            bof = False
            new_row = value
            if new_row < 0:
                new_row = 0
                bof = True
            if new_row >= len(self._dataset):
                new_row = len(self._dataset) - 1
                eof = True
            self._eof = eof
            self._bof = bof
            if old_row != new_row:
                self._cur_row = new_row
                self._do_after_scroll()
            elif (eof or bof) and self.is_new() and self.record_count() == 1:
                self._do_before_scroll()
                self._do_after_scroll()

    def _set_rec_no(self, value):
        if self._active:
            if self.filter_active():
                self.search_record(value, 0)
            else:
                self.skip(value)

    def _get_rec_no(self):
        if self._active:
            return self._cur_row

    rec_no = property (_get_rec_no, _set_rec_no)

    def filter_active(self):
        if self.on_filter_record and self.filtered:
            return True

    def first(self):
        if self.filter_active():
            self.find_first()
        else:
            self.rec_no = 0

    def last(self):
        if self.filter_active():
            self.find_last()
        else:
            self.rec_no = len(self._dataset)

    def next(self):
        if self.filter_active():
            self.find_next()
        else:
            self.rec_no += 1

    def prior(self):
        if self.filter_active():
            self.find_prior()
        else:
            self.rec_no -= 1

    def eof(self):
        return self._eof

    def bof(self):
        return self._bof

    def valid_record(self):
        if self.on_filter_record and self.filtered:
            return self.on_filter_record(self)
        else:
            return True

    def search_record(self, start, direction=1):

        def update_position():
            if self.record_count() != 0:
                self._eof = False
                self._bof = False
                if self._cur_row < 0:
                    self._cur_row = 0
                    self._bof = True
                if self._cur_row >= len(self._dataset):
                    self._cur_row = len(self._dataset) - 1
                    self._eof = True

        def check_record():
            if direction == 1:
                return self.eof()
            else:
                return self.bof()

        if self.active:
            if self.record_count() == 0:
                self.skip(start)
                return
            cur_row = self._cur_row
            self._cur_row = start + direction
            update_position()
            if direction == 0:
                if self.valid_record():
                    self._cur_row = cur_row
                    self.skip(start)
                    return
                direction = 1
            while not check_record():
                if self.valid_record():
                    if start != self._cur_row:
                        row = self._cur_row
                        self._cur_row = start
                        self.skip(row)
                        return
                else:
                    self._cur_row = self._cur_row + direction
                    update_position()

    def find_first(self):
        self.search_record(-1, 1)

    def find_last(self):
        self.search_record(len(self._dataset), -1)

    def find_next(self):
        self.search_record(self.rec_no, 1)

    def find_prior(self):
        self.search_record(self.rec_no, -1)

    def record_count(self):
        if self._dataset:
            return len(self._dataset)
        else:
            return 0

    def update_controls(self, state):
        pass

    def do_internal_open(self):
        pass

    def find_rec_info(self, rec_no=None, record=None):
        if record is None:
            if rec_no is None:
                rec_no = self.rec_no
            if self.record_count() > 0:
                record = self._dataset[rec_no];
        if record and self._record_info_index > 0:
            if len(record) < self._record_info_index + 1:
                record.append([None, {}, None])
            return record[self._record_info_index]

    def get_rec_info(self, rec_no=None, record=None):
        return self.find_rec_info(rec_no, record)

    def get_records_status(self):
        info = self.get_rec_info()
        if info:
            return info[common.REC_STATUS]

    def _set_record_status(self, value):
        info = self.get_rec_info()
        if info and (self.log_changes or self._is_delta):
            info[common.REC_STATUS] = value

    record_status = property (get_records_status, _set_record_status)

    def _get_rec_change_id(self):
        info = self.get_rec_info()
        if info:
            return info[common.REC_CHANGE_ID]

    def _set_rec_change_id(self, value):
        info = self.get_rec_info()
        if info:
            info[common.REC_CHANGE_ID] = value

    rec_change_id = property (_get_rec_change_id, _set_rec_change_id)

    def rec_controls_info(self):
        info = self.get_rec_info()
        if info:
            return info[common.REC_CONTROLS_INFO]

    def _bind_fields(self, expanded=True):
        for field in self.fields:
            field.bind_index = None
            field.lookup_index = None
        j = 0
        for field in self.fields:
            if not field.master_field:
                field.bind_index = j
                j += 1
        for field in self.fields:
            if field.master_field:
                field.bind_index = field.master_field.bind_index
        self._record_lookup_index = j
        if expanded:
            for field in self.fields:
                if field.lookup_item:
                    field.lookup_index = j
                    j += 1
        self._record_info_index = j

    def get_where_list(self, field_dict):
        result = []
        for field_arg in field_dict.keys():
            field_name = field_arg
            pos = field_name.find('__')
            if pos > -1:
                filter_str = field_name[pos+2:]
                field_name = field_name[:pos]
            else:
                filter_str = 'eq'
            filter_type = common.FILTER_STR.index(filter_str)
            if filter_type != -1:
                filter_type += 1
            else:
                raise RuntimeError('%s: set_where method arument error %s' % (self.item_name, field_arg))
            field = self._field_by_name(field_name)
            if not field:
                raise RuntimeError('%s: set_where method arument error %s: ' % (self.item_name, field_arg))
            value = field_dict[field_arg]
            if not value is None:
                result.append((field_name, filter_type, value))
        return result

    def set_fields(self, *fields):
        self._select_field_list = [field for field in fields];

    def set_where(self, **fields):
        self._where_list = self.get_where_list(fields)

    def get_order_by_list(self, field_list):
        result = []
        for field in field_list:
            field_name = field
            desc = False
            if field[0] == '-':
                desc = True
                field_name = field[1:]
            try:
                fld = self._field_by_name(field_name)
            except:
                raise RuntimeError('%s: change_order method arument error - %s' % (self.item_name, field))
            result.append([fld.ID, desc])
        return result

    def set_order_by(self, *fields):
        self._order_by_list = self.get_order_by_list(fields)

    def _update_fields(self, fields):

        for field in self.fields:
            if hasattr(self, field.field_name):
                delattr(self, field.field_name)
        if fields:
            self.fields = []
            for field_name in fields:
                field = self._field_by_name(field_name)
                if field:
                    self.fields.append(field)
                else:
                    raise Exception, '%s - _do_before_open method error: there is no field with field_name: %s' % (self.item_name, field_name)
        else:
            self.fields = list(self._fields)
        for field in self.fields:
            if not hasattr(self, field.field_name):
                setattr(self, field.field_name, field)
        for sys_field_name in ['_primary_key', '_deleted_flag', '_master_id', '_master_rec_id']:
            sys_field = getattr(self, sys_field_name)
            if sys_field:
                field = self.field_by_name(sys_field)
                if field:
                    setattr(self, sys_field_name + '_field', field)

    def _do_before_open(self, expanded, fields, where, order_by, open_empty,
        params, offset, limit, funcs, group_by):

        if self.on_before_open:
             self.on_before_open(self, params)

        result = None
        params['__expanded'] = expanded
        params['__fields'] = []
        params['__filters'] = []
        filters = []

        if fields is None and self._select_field_list:
            fields = self._select_field_list
        self._update_fields(fields)
        if fields:
            params['__fields'] = fields
        if not open_empty:
            params['__limit'] = 0
            params['__offset'] = 0
            if limit:
                params['__limit'] = limit
                if offset:
                    params['__offset'] = offset
            if where:
                filters = self.get_where_list(where)
            elif self._where_list:
                filters = list(self._where_list)
            else:
                if self.filters:
                    for fltr in self.filters:
                        if not fltr.value is None:
                            filters.append([fltr.field.field_name, fltr.filter_type, fltr.value])
            if params.get('__search'):
                field_name, text = params['__search']
                filters.append([field_name, common.FILTER_CONTAINS_ALL, text])
            params['__filters'] = filters
            if order_by:
                params['__order'] = self.get_order_by_list(order_by)
            elif self._order_by_list:
                params['__order'] = list(self._order_by_list)
            elif self._order_by:
                params['__order'] = self._order_by
            if funcs:
                params['__funcs'] = funcs
            if group_by:
                params['__group_by'] = group_by
            self._order_by_list = []
            self._where_list = []
            self._open_params = params

    def _do_after_open(self):
        if self.on_after_open:
            self.on_after_open(self)

    def open(self, expanded, fields, where, order_by, open_empty, params,
        offset, limit, funcs, group_by):
        if not params:
            params = {}
        self._do_before_open(expanded, fields, where, order_by, open_empty,
            params, offset, limit, funcs, group_by)
        self.change_log.prepare()
        self._bind_fields(expanded)
        self._dataset = []
        if not open_empty:
            self.do_open(params)
        self._active = True
        self._cur_row = None
        self.item_state = common.STATE_BROWSE
        self.first()
        self._do_after_open()
        self.update_controls(common.UPDATE_OPEN)
        if self.on_filters_applied:
            self.on_filters_applied(self)

    def do_open(self, params=None):
        if not params:
            params = self._open_params
        rows, error_mes = self.do_internal_open(params)
        if error_mes:
            raise RuntimeError(error_mes)
        else:
            self._dataset = rows

    def close(self):
        self._active = False
        self._cur_row = None
        self._dataset = None
        self.close_details()

    def close_details(self):
        for detail in self.details:
            detail.close()

    def new_record(self):
        result = [None for field in self.fields if not field.master_field]
        if self.expanded:
            result += [None for field in self.fields if field.lookup_item]
        return result

    def _do_before_append(self):
        if self.on_before_append:
            self.on_before_append(self)

    def _do_after_append(self):
        for field in self.fields:
            if field.default_value:
                try:
                    field.text = field.default_value
                except:
                    pass
        self._modified = False
        if self.on_after_append:
            self.on_after_append(self)

    def append(self):
        if not self.active:
            raise DatasetException(u"Can't insert record in %s: %s is not active" % (self.item_name, self.item_name))
        if self.item_state != common.STATE_BROWSE:
            raise DatasetException(u"Can't insert record in %s: %s is not in browse state" % (self.item_name, self.item_name))
        self._do_before_append()
        self._do_before_scroll()
        self._old_row = self.rec_no
        self.item_state = common.STATE_INSERT
        self._dataset.append(self.new_record())
        self._cur_row = len(self._dataset) - 1
#        self._modified = False
        self.record_status = common.RECORD_INSERTED
        self.update_controls(common.UPDATE_APPEND)
        self._do_after_scroll()
        self._do_after_append()

    def insert(self):
        if not self.active:
            raise DatasetException(u"Can't insert record in %s: %s is not active" % (self.item_name, self.item_name))
        if self.item_state != common.STATE_BROWSE:
            raise DatasetException(u"Can't insert record in %s: %s is not in browse state" % (self.item_name, self.item_name))
        self._do_before_append()
        self._do_before_scroll()
        self._old_row = self.rec_no
        self.item_state = common.STATE_INSERT
        self._dataset.insert(0, self.new_record())
        self._cur_row = 0
        self._modified = False
        self.record_status = common.RECORD_INSERTED
        self.update_controls(common.UPDATE_INSERT)
        self._do_after_scroll()
        self._do_after_append()

    def rec_inserted(self):
        return self.record_status == common.RECORD_INSERTED

    def rec_deleted(self):
        return self.record_status == common.RECORD_DELETED

    def rec_modified(self):
        return self.record_status in (common.RECORD_MODIFIED, common.RECORD_DETAILS_MODIFIED)

    def is_browsing(self):
        return self.item_state == consts.STATE_BROWSE

    def is_changing(self):
        return (self.item_state == common.STATE_INSERT) or (self.item_state == common.STATE_EDIT)

    def is_new(self):
        return self.item_state == common.STATE_INSERT

    def is_edited(self):
        return self.item_state == common.STATE_EDIT

    def is_deleting(self):
        return self.item_state == common.STATE_DELETE

    def _do_before_edit(self):
        if self.on_before_edit:
            self.on_before_edit(self)

    def _do_after_edit(self):
        if self.on_after_edit:
            self.on_after_edit(self)

    def edit(self):
        if not self.active:
            raise DatasetException(u"Can't edit record in %s: %s is not active" % (self.item_name, self.item_name))
        if self.item_state != common.STATE_BROWSE:
            raise DatasetException(u"Can't edit record in %s: %s is not in browse state" % (self.item_name, self.item_name))
        if self.record_count() == 0:
            raise DatasetException(u"Can't edit record in %s: %s' record list is empty" % (self.item_name, self.item_name))
        self._do_before_edit()
        self.change_log.log_change()
        self._buffer = self.change_log.store_record_log()
        self.item_state = common.STATE_EDIT
        self._old_row = self.rec_no
        self._old_status = self.record_status
        self._modified = False
        self._do_after_edit()

    def _do_before_delete(self):
        if self.on_before_delete:
            self.on_before_delete(self)

    def _do_after_delete(self):
        if self.on_after_delete:
            self.on_after_delete(self)

    def delete(self):
        if not self.active:
            raise DatasetException(u"Can't delete record in %s: %s is not active" % (self.item_name, self.item_name))
        if self.record_count() == 0:
            raise DatasetException(u"Can't delete record in %s: %s' record list is empty" % (self.item_name, self.item_name))
        self.item_state = common.STATE_DELETE
        try:
            self._do_before_delete()
            self._do_before_scroll()
            self.update_controls(common.UPDATE_DELETE)
            self.change_log.log_change()
            if self.master:
                self.master._set_modified(True)
            self._dataset.remove(self._dataset[self.rec_no])
            self.rec_no = self.rec_no
            self._do_after_scroll()
            self.item_state = common.STATE_BROWSE
            self._do_after_delete()
        finally:
            self.item_state = common.STATE_BROWSE

    def _do_before_cancel(self):
        if self.on_before_cancel:
            self.on_before_cancel(self)

    def _do_after_cancel(self):
        if self.on_after_cancel:
            self.on_after_cancel(self)

    def cancel(self):
        self._do_before_cancel()
        if self.item_state == common.STATE_EDIT:
            self.change_log.restore_record_log(self._buffer)
            self.update_controls(common.UPDATE_CANCEL)
            for detail in self.details:
                detail.update_controls(common.UPDATE_OPEN)
        elif self.item_state == common.STATE_INSERT:
            self.change_log.remove_record_log()
            self.update_controls(common.UPDATE_DELETE)
            del self._dataset[self.rec_no]
        else:
            raise Exception, '%s cancel error: invalid item state' % self.item_name
        prev_state = self.item_state
        self.item_state = common.STATE_BROWSE
        if prev_state in [common.STATE_INSERT]:
            self._do_before_scroll()
        self._cur_row = self._old_row
        if prev_state in [common.STATE_EDIT]:
            self.record_status = self._old_status
        self._set_modified(False)
        if prev_state in [common.STATE_INSERT]:
            self._do_after_scroll()
        self._do_after_cancel()

    def _do_before_post(self):
        if self.on_before_post:
            self.on_before_post(self)

    def _do_after_post(self):
        if self.on_after_post:
            self.on_after_post(self)

    def post(self):
        if not self.is_changing():
            raise DatasetException, u'%s: dataset is not in edit or insert mode' % self.item_name
        self.check_record_valid()
        self._do_before_post()
        if self.master:
            self.field_by_name(self._master_id).value = self.master.ID
        for detail in self.details:
            if detail.is_changing():
                detail.post()
        if self.is_modified() or self.is_new():
            self.change_log.log_change()
        elif self.record_status == common.RECORD_UNCHANGED:
            self.change_log.remove_record_log()
        self._modified = False
        self.item_state = common.STATE_BROWSE
        self._do_after_post()
        if not self.valid_record():
            self.update_controls(common.UPDATE_DELETE)
            self.search_record(self.rec_no, 0)

    def check_record_valid(self):
        for field in self.fields:
            field.check_valid()
        return True

    def locate(self, fields, values):
        clone = self.clone()

        def record_found():
            if isinstance(fields, (list, tuple)):
                for i, field in enumerate(fields):
                    if clone.field_by_name(field).value != values[i]:
                        return False
                return True
            else:
                if clone.field_by_name(fields).value == values:
                    return True

        for c in clone:
            if record_found():
                self.rec_no = clone.rec_no
                return True

    def filter_index(self, filter_name):
        filter = self.filter_by_name(filter_name)
        if filter:
            return self.filters.index(filter)

    def get_field_values(self):
        result = []
        for field in self.fields:
            result.append(field.value)
        return result

    def set_field_values(self, values):
        for i, field in enumerate(self.fields):
            field.value = values[i]

    def compare_field_values(self, values):
        result = True
        for i, field in enumerate(self.fields):
            if self.get_field_values() != values:
                result = False
                break
        return result

    def clear_filters(self):
        for filter in self.filters:
            filter.value = None

    def get_filter_values(self):
        result = []
        for filter in self.filters:
            result.append(filter.value)
        return result

    def set_filter_values(self, values):
        for i, filter in enumerate(self.filters):
            filter.value = values[i]

    def search(self, field_name, text):
        searchText = text.strip()
        params = {};
        if len(searchText):
            params['__search'] = [field_name, searchText]
            self.open(params=params)
        else:
            self.open()

    def set_filters(self, **filters):
        self.clear_filters()
        for filter_name in filters.keys():
            try:
                filter = self.filter_by_name(filter_name)
                filter.value = filters[filter_name]
            except Exception, e:
                raise RuntimeError('%s: set_filters method arument error %s=%s: %s' % (self.item_name, filter_name, filters[filter_name], e))

    def find_default_field(self):
        for field in self.fields:
            if field.is_default:
                return field

    default_field = property (find_default_field)

    def round(self, value, dec):
        return round(value, dec)

class MasterDataSet(AbstractDataSet):
    def __init__(self):
        AbstractDataSet.__init__(self)
        self.details_active = False

    def _copy(self, filters=True, details=True, handlers=True):
        result = super(MasterDataSet, self)._copy(filters, details, handlers)
        if details:
            for detail in self.details:
                copy_table = detail._copy(filters, details, handlers)
                copy_table.owner = result
                copy_table.master = result
                copy_table.expanded = detail.expanded
                result.details.append(copy_table)
                result.items.append(copy_table)
                if not hasattr(result, copy_table.item_name):
                    setattr(result, copy_table.item_name, copy_table)
                if not hasattr(result.details, copy_table.item_name):
                    setattr(result.details, copy_table.item_name, copy_table)

        return result

    def do_apply(self, params):
        pass

    def apply(self, params=None):
        result = None
        if self.is_changing():
            self.post()
        if self.on_before_apply:
            result = self.on_before_apply(self)
            if result:
                params = result
        self.do_apply(params)
        if self.on_after_apply:
            self.on_after_apply(self)

    def detail_by_ID(self, ID):
        ID = int(ID)
        for detail in self.details:
            if detail.ID == ID:
                return detail

    def delta(self, changes=None):
        if not changes:
            changes = {}
            self.change_log.get_changes(changes)
        result = self.copy(filters=False, details=True, handlers=False)
        result.log_changes = False
        result.expanded = False
        result._is_delta = True
        for detail in result.details:
            detail.log_changes = False
            detail.expanded = False
            detail._is_delta = True
        result.details_active = True
        result.change_log.set_changes(changes)
        result._dataset = result.change_log.records
        result._update_fields(result.change_log.fields)
        result._bind_fields(result.change_log.expanded)
        result.item_state = common.STATE_BROWSE
        result._cur_row = None
        result._active = True
        result.first()
        return result

    def detail_by_name(self, name):
        for detail in self.details:
            if detail.item_name == name:
                return detail

    def open_details(self):
        for detail in self.details:
            if not detail.disabled:
                detail.open()

    def _do_after_scroll(self):
        if self.details_active:
            self.open_details()
        else:
            self.close_details()
        super(MasterDataSet, self)._do_after_scroll()

    def _set_read_only(self, value):
        super(MasterDataSet, self)._set_read_only(value)
        for detail in self.details:
            detail._set_read_only(value)


class MasterDetailDataset(MasterDataSet):
    def __init__(self):
        MasterDataSet.__init__(self)
        self.disabled = False

    def find_change_log(self):
        if self.master:
            if self.master.record_status != common.RECORD_UNCHANGED:
                return self.master.change_log.get_detail_log(str(self.ID))

    def open(self, expanded=None, fields=None, where=None, order_by=None,
        open_empty=False, params=None, offset=None, limit=None, funcs=None,
        group_by=None):
        if expanded is None:
            expanded = self.expanded
        else:
            self.expanded = expanded
        if not params:
            params = {}
        if self.master:
            if not self.disabled and self.master.record_count() > 0:
                records = None
                if self.master.is_new():
                    records = []
                else:
                    log = self.find_change_log()
                    if log:
                        records = log['records']
                        fields = log['fields']
                        expanded = log['expanded']
                if not records is None:
                    self._do_before_open(expanded, fields, where, order_by,
                        open_empty, params, offset, limit, funcs, group_by)
                    self._bind_fields(expanded)
                    if self.master.is_new():
                        self.change_log.prepare()
                    self._dataset = records
                    self._active = True
                    self.item_state = common.STATE_BROWSE
                    self.first()
                    self._do_after_open()
                    self.update_controls(common.UPDATE_OPEN)
                else:
                    params['__master_id'] = self.master.ID
                    params['__master_rec_id'] = self.master.field_by_name(self.master._primary_key).value
                    return super(MasterDetailDataset, self).open(expanded,
                        fields, where, order_by, open_empty, params, offset,
                        limit, funcs, group_by)
            else:
                return
        else:
            return super(MasterDetailDataset, self).open(expanded,
                fields, where, order_by, open_empty, params, offset, limit,
                funcs, group_by)

    def open__in(self, ids, expanded=None, fields=None, where=None, order_by=None, open_empty=False, params=None, offset=None, limit=None):

        def slice_ids(ids):
            MAX_IN_LIST = 999
            result = []
            while True:
                result.append(ids[0:MAX_IN_LIST])
                ids = ids[MAX_IN_LIST:]
                if len(ids) == 0:
                    break;
            return result

        params = {}
        for args, value in locals().iteritems():
            if not args in ['self', 'ids', 'params', 'slice_ids']:
                params[args] = value
        if type(ids) == dict:
            id_field_name = ids.keys()[0]
            ids = ids[id_field_name]
        elif type(ids) == list:
            id_field_name = 'id'
        else:
            raise Exception, u'Invalid ids parameter in open_for_ids method of item "%s"' % self.item_name
        if params['where'] is None:
            params['where'] = {}
        on_after_open = self.on_after_open
        self.on_after_open = None
        on_filters_applied = self.on_filters_applied
        self.on_filters_applied = None
        records = []
        lst = slice_ids(ids)
        for l in lst:
            params['where'][id_field_name + '__in'] = l
            self.open(**params)
            records += self._dataset
        self._dataset = records
        self.first()
        self.on_after_open = on_after_open
        self.on_filters_applied = on_filters_applied
        self._do_after_open()
        self.update_controls(common.UPDATE_OPEN)
        if self.on_filters_applied:
            self.on_filters_applied(self)

    def insert(self):
        if self.master and not self.master.is_changing():
            raise DatasetException, u"%s: can't insert record - master item is not in edit or insert mode" % self.owner.item_name
        super(MasterDetailDataset, self).insert()

    def append(self):
        if self.master and not self.master.is_changing():
            raise DatasetException, u"%s: can't append record - master item is not in edit or insert mode" % self.owner.item_name
        super(MasterDetailDataset, self).append()

    def edit(self):
        if self.master and not self.master.is_changing():
            raise DatasetException, u"%s: can't edit record - master item is not in edit or insert mode" % self.owner.item_name
        super(MasterDetailDataset, self).edit()

    def delete(self):
        if self.master and not self.master.is_changing():
            raise DatasetException, u"%s: can't delete record - master item is not in edit or insert mode" % self.owner.item_name
        super(MasterDetailDataset, self).delete()

    def _set_modified(self, value):
        self._modified = value
        if self.master and value:
            self.master._set_modified(value)

    def is_modified(self):
        return super(MasterDetailDataset, self).is_modified()

#    modified = property (is_modified, _set_modified)

    def _get_read_only(self):
        if self.master and self.parent_read_only:
            return self.master.read_only
        else:
            return super(MasterDetailDataset, self)._get_read_only()

    def _set_read_only(self, value):
        return super(MasterDetailDataset, self)._set_read_only(value)

    read_only = property (_get_read_only, _set_read_only)

class Dataset(MasterDetailDataset):
    pass
